"""Class that searches indexes."""

import datetime as dt
import functools
import logging
from time import perf_counter
from typing import Any, Union, Dict, Tuple

import faiss
import numpy as np
import pandas as pd
from scipy.stats import norm

try:
    from tqdm.auto import tqdm
except:
    try:
        from tqdm import tqdm
    except:
        raise Exception("Can't import tqdm")

from ..utils.metrics import check_repeats
from ..utils.metrics import matching_quality


def timer(func):
    """Decorator to measure the execution time of a function.

    Uses time.perf_counter() to determine the start and end times
    of the decorated function and then prints the total execution time

    Usage Example:

        @timer
        def example_function():
            ...

    Args:
        func: The function whose execution time is to be measured

    Returns:
        Wrapped version of the original function with added time measurement
    """

    @functools.wraps(func)
    def _wrapper(*args, **kwargs):
        start = perf_counter()
        result = func(*args, **kwargs)
        runtime = perf_counter() - start
        print(f"{func.__name__} took {runtime:.4f} secs")
        return result

    return _wrapper


faiss.cvar.distance_compute_blas_threshold = 100000
POSTFIX = "_matched"
POSTFIX_BIAS = "_matched_bias"

logger = logging.getLogger("Faiss hypex")
console_out = logging.StreamHandler()
logging.basicConfig(
    handlers=(console_out,),
    format="[%(asctime)s | %(name)s | %(levelname)s]: %(message)s",
    datefmt="%d.%m.%Y %H:%M:%S",
    level=logging.INFO,
)


class FaissMatcher:
    """A class used to match instances using Faiss library."""

    def __init__(
        self,
        df: pd.DataFrame,
        outcomes: str,
        treatment: str,
        info_col: list,
        features: [list, pd.DataFrame] = None,
        group_col: Union[str, list] = None,
        weights: dict = None,
        sigma: float = 1.96,
        validation: bool = None,
        refuter: str = "random_feature",
        n_neighbors: int = 10,
        silent: bool = True,
        pbar: bool = True,
    ):
        """Construct all the necessary attributes.

        Args:
            df:
                The input dataframe
            outcomes:
                The target column name
            treatment:
                The column name with treatment
            info_col:
                A list with informational column names
            features:
                A list with names of feature using to matching. Defaults to None
            group_col:
                The column for stratification. Defaults to None
            weights:
                Dict with wight of features to matching. If you would like that matching will be more for
                1 feature and less for another one
            sigma:
                The significant level for confidence interval calculation Defaults to 1.96
            validation:
                The flag for validation of estimated ATE with default method `random_feature`
            n_neighbors:
                The number of neighbors to find for each object. Defaults to 10
            silent:
                Write logs in debug mode
            pbar:
                Display progress bar while get index
        """
        self.n_neighbors = n_neighbors
        if group_col is None:
            self.df = df
        else:
            group_col = group_col if isinstance(group_col, list) else [group_col]
            self.df = df.sort_values([treatment, group_col[0]])
        self.columns_del = [outcomes]
        if info_col:
            self.info_col = info_col
        else:
            self.info_col = []

        if self.info_col is not None:
            self.columns_del = self.columns_del + [
                x for x in self.info_col if x in self.df.columns
            ]
        self.outcomes = outcomes if type(outcomes) == list else [outcomes]
        self.treatment = treatment

        if features is None:
            self.columns_match = list(
                set(
                    [x for x in list(self.df.columns) if x not in self.info_col]
                    + [self.treatment]
                    + self.outcomes
                )
            )
        else:
            try:
                self.columns_match = (
                    features["Feature"].tolist() + [self.treatment] + self.outcomes
                )
            except TypeError:
                self.columns_match = features + [self.treatment] + self.outcomes

        self.features_quality = (
            self.df.drop(columns=[self.treatment] + self.outcomes + self.info_col)
            .select_dtypes(
                include=["int16", "int32", "int64", "float16", "float32", "float64"]
            )
            .columns
        )
        self.dict_outcome_untreated = {}
        self.dict_outcome_treated = {}
        self.group_col = group_col
        self.weights = weights
        self.treated_index = None
        self.untreated_index = None
        self.orig_treated_index = None
        self.orig_untreated_index = None
        self.results = {}
        self.ATE = None
        self.refuter = refuter
        self.delta_t = None
        self.sigma = sigma
        self.quality_dict = {}
        self.rep_dict = None
        self.validation = validation
        self.silent = silent
        self.pbar = pbar
        self.tqdm = None
        self.results = pd.DataFrame()

    def __getstate__(self) -> dict:
        """Prepare the object for serialization.

        This method is called when the object is about to be serialized.
        It removes the `tqdm` attribute from the object's dictionary
        because `tqdm` objects cannot be serialized.

        Returns:
            A copy of the object's dictionary with the `tqdm` attribute removed.
        """
        state = self.__dict__.copy()
        if "tqdm" in state:
            del state["tqdm"]
        return state

    def __setstate__(self, state: dict):
        """Restore the object after deserialization.

        This method is called when the object is deserialized.
        It adds the `tqdm` attribute back to the object's dictionary
        if the `pbar` attribute is True.

        Args:
            state:
                The deserialized state of the object
        """
        if "pbar" in state and state["pbar"]:
            state["tqdm"] = None
        self.__dict__.update(state)

    def _get_split(self, df: pd.DataFrame) -> (pd.DataFrame, pd.DataFrame):
        """Creates split data by treatment column.

        Separate treatment column with 1 (treated) an 0 (untreated),
        scales and transforms treatment column

        Args:
            df:
                The input dataframe

        Returns:
            Tuple of dataframes - one for treated (df[self.treatment] == 1]) and
            one for untreated (df[self.treatment] == 0]). Drops self.outcomes and
            `self.treatment` columns

        """
        logger.debug("Creating split data by treatment column")

        treated = df[df[self.treatment] == 1].drop(
            [self.treatment] + self.outcomes, axis=1
        )
        untreated = df[df[self.treatment] == 0].drop(
            [self.treatment] + self.outcomes, axis=1
        )

        return treated, untreated

    def _predict_outcome(self, std_treated: pd.DataFrame, std_untreated: pd.DataFrame):
        """Applies LinearRegression to input arrays.

        Calculate biases of treated and untreated values,
        creates dict of y - regular, matched and without bias.

        Args:
            std_treated:
                The dataframe of treated data
            std_untreated:
                The dataframe of untreated data

        """
        logger.debug("Predicting target by Linear Regression")

        start_time = dt.datetime.now()
        logger.debug("start --")

        self.dict_outcome_untreated = {}
        self.dict_outcome_treated = {}
        df = self.df.drop(columns=self.info_col)

        for outcome in self.outcomes:
            y_untreated = df[df[self.treatment] == 0][outcome].to_numpy()
            y_treated = df[df[self.treatment] == 1][outcome].to_numpy()

            x_treated = std_treated.to_numpy()
            x_untreated = std_untreated.to_numpy()
            y_match_treated = np.array(
                [y_untreated[idx].mean() for idx in self.treated_index]
            )
            y_match_untreated = np.array(
                [y_treated[idx].mean() for idx in self.untreated_index]
            )
            x_match_treated = np.array(
                [x_untreated[idx].mean(0) for idx in self.treated_index]
            )
            x_match_untreated = np.array(
                [x_treated[idx].mean(0) for idx in self.untreated_index]
            )
            bias_coefs_c = bias_coefs(self.untreated_index, y_treated, x_treated)
            bias_coefs_t = bias_coefs(self.treated_index, y_untreated, x_untreated)
            bias_c = bias(x_untreated, x_match_untreated, bias_coefs_c)
            bias_t = bias(x_treated, x_match_treated, bias_coefs_t)

            y_match_treated_bias = y_treated - y_match_treated + bias_t
            y_match_untreated_bias = y_match_untreated - y_untreated - bias_c

            self.delta_t = round(
                abs(bias_t.mean()) * 100 / abs(y_match_treated_bias.mean()), 1
            )

            self.dict_outcome_untreated[outcome] = y_untreated
            self.dict_outcome_untreated[outcome + POSTFIX] = y_match_untreated
            self.dict_outcome_untreated[outcome + POSTFIX_BIAS] = y_match_untreated_bias

            self.dict_outcome_treated[outcome] = y_treated
            self.dict_outcome_treated[outcome + POSTFIX] = y_match_treated
            self.dict_outcome_treated[outcome + POSTFIX_BIAS] = y_match_treated_bias

        end_time = dt.datetime.now()
        total = dt.datetime.strptime(
            str(end_time - start_time), "%H:%M:%S.%f"
        ).strftime("%H:%M:%S")
        logger.debug(f"end -- [work time{total}]")

    def _create_outcome_matched_df(
        self, dict_outcome: dict, is_treated: bool
    ) -> pd.DataFrame:
        """Creates dataframe with outcomes values and treatment.

        Args:
            dict_outcome:
                A dictionary containing outcomes
            is_treated:
                A boolean value indicating whether the outcome is treated or not

        Returns:
            A dataframe with matched outcome and treatment columns

        """
        df_pred = pd.DataFrame(dict_outcome)
        df_pred[self.treatment] = int(is_treated)
        df_pred[self.treatment + POSTFIX] = int(not is_treated)

        return df_pred

    def _create_features_matched_df(
        self, index: np.ndarray, is_treated: bool
    ) -> pd.DataFrame:
        """Creates matched dataframe with features.

        Args:
            index:
                An array of indices
            is_treated:
                A boolean value indicating whether the outcome is treated or not


        Returns:
            A dataframe of matched features

        """
        df = self.df.drop(columns=self.outcomes + self.info_col)

        if self.group_col is None:
            untreated_index = df[
                df[self.treatment] == int(not is_treated)
            ].index.to_numpy()
            converted_index = [untreated_index[i] for i in index]
            filtered = df.loc[df[self.treatment] == int(not is_treated)].values
            untreated_df = pd.DataFrame(
                data=np.array([filtered[idx].mean(axis=0) for idx in index]),
                columns=df.columns,
            )  # добавить дату в данные и пофиксить баги с этим (тут ломалось)
            if self.info_col is not None and len(self.info_col) != 1:
                untreated_df["index"] = pd.Series(converted_index)
                treated_df = df[df[self.treatment] == int(is_treated)].reset_index()
            else:
                ids = self.df[df[self.treatment] == int(not is_treated)][
                    self.info_col
                ].values.ravel()
                converted_index = [ids[i] for i in index]
                untreated_df["index"] = pd.Series(converted_index)
                treated_df = df[df[self.treatment] == int(is_treated)].reset_index()
                treated_df["index"] = self.df[
                    self.df[self.treatment] == int(is_treated)
                ][self.info_col].values.ravel()
        else:
            df = df.sort_values([self.treatment, self.group_col[0]])
            untreated_index = df[
                df[self.treatment] == int(not is_treated)
            ].index.to_numpy()
            converted_index = [untreated_index[i] for i in index]
            filtered = df.loc[df[self.treatment] == int(not is_treated)]
            cols_untreated = [
                col for col in filtered.columns if col != self.group_col[0]
            ]
            filtered = filtered.drop(columns=self.group_col[0]).to_numpy()
            untreated_df = pd.DataFrame(
                data=np.array([filtered[idx].mean(axis=0) for idx in index]),
                columns=cols_untreated,
            )
            treated_df = df[df[self.treatment] == int(is_treated)].reset_index()
            grp = treated_df[self.group_col[0]]
            untreated_df[self.group_col[0]] = grp
            if self.info_col is not None and len(self.info_col) != 1:
                untreated_df["index"] = pd.Series(converted_index)
            else:
                ids = (
                    self.df[df[self.treatment] == int(not is_treated)]
                    .sort_values([self.treatment, self.group_col[0]])[self.info_col]
                    .values.ravel()
                )
                converted_index = [ids[i] for i in index]
                untreated_df["index"] = pd.Series(converted_index)
                treated_df["index"] = self.df[
                    self.df[self.treatment] == int(is_treated)
                ][self.info_col].values.ravel()
        untreated_df.columns = [col + POSTFIX for col in untreated_df.columns]

        x = pd.concat([treated_df, untreated_df], axis=1).drop(
            columns=[self.treatment, self.treatment + POSTFIX], axis=1
        )
        return x

    def _create_matched_df(self) -> pd.DataFrame:
        """Creates matched df of features and outcome.

        Returns:
            Matched dataframe
        """
        df_pred_treated = self._create_outcome_matched_df(
            self.dict_outcome_treated, True
        )
        df_pred_untreated = self._create_outcome_matched_df(
            self.dict_outcome_untreated, False
        )

        df_matched = pd.concat([df_pred_treated, df_pred_untreated])

        treated_x = self._create_features_matched_df(self.treated_index, True)
        untreated_x = self._create_features_matched_df(self.untreated_index, False)

        untreated_x = pd.concat([treated_x, untreated_x])

        columns = list(untreated_x.columns) + list(df_matched.columns)

        df_matched = pd.concat([untreated_x, df_matched], axis=1, ignore_index=True)
        df_matched.columns = columns

        return df_matched

    def calc_atc(self, df: pd.DataFrame, outcome: str) -> tuple:
        """Calculates Average Treatment Effect for the control group (ATC).

        Effect on control group if it was affected

        Args:
            df:
                Input dataframe
            outcome:
                The outcome to be considered for treatment effect

        Returns:
            Contains ATC, scaled counts, and variances as numpy arrays

        """
        logger.debug("Calculating ATC")

        df = df[df[self.treatment] == 0]
        N_c = len(df)
        ITT_c = df[outcome + POSTFIX_BIAS]
        scaled_counts_c = scaled_counts(N_c, self.treated_index, self.silent)

        vars_c = np.repeat(ITT_c.var(), N_c)  # conservative
        atc = ITT_c.mean()

        return atc, scaled_counts_c, vars_c

    def calc_att(self, df: pd.DataFrame, outcome: str) -> tuple:
        """Calculates Average Treatment Effect for the treated (ATT).

        Args:
            df:
                Input dataframe
            outcome:
                The outcome to be considered for treatment effect

        Returns:
            Contains ATT, scaled counts, and variances as numpy arrays

        """
        logger.debug("Calculating ATT")

        df = df[df[self.treatment] == 1]
        N_t = len(df)
        ITT_t = df[outcome + POSTFIX_BIAS]
        scaled_counts_t = scaled_counts(N_t, self.untreated_index, self.silent)

        vars_t = np.repeat(ITT_t.var(), N_t)  # conservative
        att = ITT_t.mean()

        return att, scaled_counts_t, vars_t

    def _calculate_ate_all_target(self, df: pd.DataFrame):
        """Creates dictionaries of all effect: ATE, ATC, ATT.

        Args:
            df:
                Input dataframe

        """
        logger.debug("Creating dicts of all effects: ATE, ATC, ATT")

        att_dict = {}
        atc_dict = {}
        ate_dict = {}
        N = len(df)
        N_t = df[self.treatment].sum()
        N_c = N - N_t

        for outcome in self.outcomes:
            att, scaled_counts_t, vars_t = self.calc_att(df, outcome)
            atc, scaled_counts_c, vars_c = self.calc_atc(df, outcome)
            ate = (N_c / N) * atc + (N_t / N) * att

            att_se = calc_att_se(vars_c, vars_t, scaled_counts_c)
            atc_se = calc_atc_se(vars_c, vars_t, scaled_counts_t)
            ate_se = calc_ate_se(vars_c, vars_t, scaled_counts_c, scaled_counts_t)

            ate_dict[outcome] = [
                ate,
                ate_se,
                pval_calc(ate / ate_se),
                ate - self.sigma * ate_se,
                ate + self.sigma * ate_se,
            ]
            atc_dict[outcome] = [
                atc,
                atc_se,
                pval_calc(atc / atc_se),
                atc - self.sigma * atc_se,
                atc + self.sigma * atc_se,
            ]
            att_dict[outcome] = [
                att,
                att_se,
                pval_calc(att / att_se),
                att - self.sigma * att_se,
                att + self.sigma * att_se,
            ]

        self.ATE, self.ATC, self.ATT = ate_dict, atc_dict, att_dict
        self.val_dict = ate_dict

    def matching_quality(self, df_matched) -> Dict[str, Union[Dict[str, float], float]]:
        """Estimated the quality of covariates balance and repeat fraction.

        Calculates population stability index,Standardized mean difference
        and Kolmogorov-Smirnov test for numeric values. Returns a dictionary of reports.

        Args:
            df_matched:
                Matched DataFrame to calculate quality

        Returns:
            dictionary containing PSI, KS-test, SMD data and repeat fractions

        """
        if self.silent:
            logger.debug("Estimating quality of matching")
        else:
            logger.info("Estimating quality of matching")

        psi_columns = set(self.columns_match)
        psi_columns = list(psi_columns - set([self.treatment] + self.outcomes))
        psi_data, ks_data, smd_data = matching_quality(
            df_matched,
            self.treatment,
            sorted(self.features_quality),
            sorted(psi_columns),
            self.silent,
        )

        rep_dict = {
            "match_control_to_treat": check_repeats(
                np.concatenate(self.treated_index), silent=self.silent
            ),
            "match_treat_to_control": check_repeats(
                np.concatenate(self.untreated_index), silent=self.silent
            ),
        }

        self.quality_dict = {
            "psi": psi_data,
            "ks_test": ks_data,
            "smd": smd_data,
            "repeats": rep_dict,
        }

        rep_df = pd.DataFrame.from_dict(rep_dict, orient="index").rename(
            columns={0: "value"}
        )
        self.rep_dict = rep_df

        if self.silent:
            logger.debug(f"PSI info: \n {psi_data.head(10)} \nshape:{psi_data.shape}")
            logger.debug(
                f"Kolmogorov-Smirnov test info: \n {ks_data.head(10)} \nshape:{ks_data.shape}"
            )
            logger.debug(
                f"Standardised mean difference info: \n {smd_data.head(10)} \nshape:{smd_data.shape}"
            )
            logger.debug(f"Repeats info: \n {rep_df.head(10)}")
        else:
            logger.info(f"PSI info: \n {psi_data.head(10)} \nshape:{psi_data.shape}")
            logger.info(
                f"Kolmogorov-Smirnov test info: \n {ks_data.head(10)} \nshape:{ks_data.shape}"
            )
            logger.info(
                f"Standardised mean difference info: \n {smd_data.head(10)} \nshape:{smd_data.shape}"
            )
            logger.info(f"Repeats info: \n {rep_df.head(10)}")

        return self.quality_dict

    def _group_match_attribute(self) -> pd.DataFrame:
        """Modifies the dataframe, removing categories for grouping for which it
         is impossible to perform a Cholesky decomposition.

        Returns:
            Dataframe with deleted categories.

        """
        df = self.df.drop(columns=self.info_col)
        groups = sorted(df[self.group_col[0]].unique())
        group_error = []

        for group in groups:
            df_group = df[df[self.group_col] == group]
            temp = df_group[self.columns_match + self.group_col]
            temp = temp.loc[:, (temp != 0).any(axis=0)].drop(columns=self.group_col)

            treated, untreated = self._get_split(temp)

            xc = untreated.to_numpy()
            xt = treated.to_numpy()

            cov_c = np.cov(xc, rowvar=False, ddof=0)
            cov_t = np.cov(xt, rowvar=False, ddof=0)
            cov = (cov_c + cov_t) / 2

            epsilon = 1e-3
            cov = cov + np.eye(cov.shape[0]) * epsilon
            try:
                np.linalg.cholesky(cov)
            except:
                if not self.validation:
                    logger.info(f"Grouping by attribute {group} is not possible")
                group_error.append(group)

                treated_duplicated = treated.columns[treated.T.duplicated()].values
                untreated_duplicated = untreated.columns[
                    untreated.T.duplicated()
                ].values
                if treated_duplicated is not [] and not self.validation:
                    logger.info(
                        f"The data has the duplicate columns: {treated_duplicated} in test group"
                    )
                if untreated_duplicated is not [] and not self.validation:
                    logger.info(
                        f"The data has the duplicate columns: {untreated_duplicated} in control group"
                    )

        df_group_positive = self.df[~self.df[self.group_col[0]].isin(group_error)]

        return df_group_positive

    def group_match(self):
        """Matches the dataframe if it divided by groups.

        Returns:
            A tuple containing the matched dataframe and metrics such as ATE, ATT and ATC

        """
        self.df = self._group_match_attribute()
        df = self.df.drop(columns=self.info_col)
        groups = sorted(df[self.group_col[0]].unique())
        matches_c = []
        matches_t = []
        group_arr_c = df[df[self.treatment] == 0][self.group_col[0]].to_numpy()
        group_arr_t = df[df[self.treatment] == 1][self.group_col[0]].to_numpy()
        treat_arr_c = df[df[self.treatment] == 0][self.treatment].to_numpy()
        treat_arr_t = df[df[self.treatment] == 1][self.treatment].to_numpy()

        if self.pbar:
            self.tqdm = tqdm(total=len(groups) * 2)

        for group in groups:
            df_group = df[df[self.group_col[0]] == group]
            temp = df_group[self.columns_match + [self.group_col[0]]]
            temp = temp.loc[:, (temp != 0).any(axis=0)].drop(columns=self.group_col[0])
            treated, untreated = self._get_split(temp)

            std_treated_np, std_untreated_np = _transform_to_np(
                treated, untreated, self.weights, self.validation
            )

            if self.pbar:
                self.tqdm.set_description(desc=f"Get untreated index by group {group}")
            matches_u_i = _get_index(std_treated_np, std_untreated_np, self.n_neighbors)

            if self.pbar:
                self.tqdm.update(1)
                self.tqdm.set_description(desc=f"Get treated index by group {group}")
            matches_t_i = _get_index(std_untreated_np, std_treated_np, self.n_neighbors)
            if self.pbar:
                self.tqdm.update(1)
                self.tqdm.refresh()

            group_mask_c = group_arr_c == group
            group_mask_t = group_arr_t == group
            matches_c_mask = np.arange(treat_arr_t.shape[0])[group_mask_t]
            matches_u_i = [matches_c_mask[i] for i in matches_u_i]
            matches_t_mask = np.arange(treat_arr_c.shape[0])[group_mask_c]
            matches_t_i = [matches_t_mask[i] for i in matches_t_i]
            matches_c.extend(matches_u_i)
            matches_t.extend(matches_t_i)

        if self.pbar:
            self.tqdm.close()

        self.untreated_index = matches_c
        self.treated_index = matches_t

        df_group = df[self.columns_match].drop(columns=self.group_col[0])
        treated, untreated = self._get_split(df_group)
        self._predict_outcome(treated, untreated)
        df_matched = self._create_matched_df()
        self._calculate_ate_all_target(df_matched)

        if self.validation:
            if self.refuter == "emissions":
                return self.report_view()
            else:
                return self.val_dict

        return self.report_view(), df_matched

    def match(self):
        """Matches the dataframe.

        Returns:
            A tuple containing the matched dataframe and metrics such as ATE, ATT and ATC

        """
        if self.group_col is not None:
            return self.group_match()

        df = self.df[self.columns_match]
        treated, untreated = self._get_split(df)

        std_treated_np, std_untreated_np = _transform_to_np(
            treated, untreated, self.weights, self.validation
        )

        if self.pbar:
            self.tqdm = tqdm(total=len(std_treated_np) + len(std_untreated_np))
            self.tqdm.set_description(desc="Get untreated index")

        untreated_index = _get_index(std_treated_np, std_untreated_np, self.n_neighbors)

        if self.pbar:
            self.tqdm.update(len(std_treated_np))
            self.tqdm.set_description(desc="Get treated index")
        treated_index = _get_index(std_untreated_np, std_treated_np, self.n_neighbors)

        if self.pbar:
            self.tqdm.update(len(std_untreated_np))
            self.tqdm.refresh()
            self.tqdm.close()

        self.untreated_index = untreated_index
        self.treated_index = treated_index

        self._predict_outcome(treated, untreated)

        df_matched = self._create_matched_df()
        self._calculate_ate_all_target(df_matched)

        if self.validation:
            if self.refuter == "emissions":
                return self.report_view()
            else:
                return self.val_dict

        return self.report_view(), df_matched

    def report_view(self) -> pd.DataFrame:
        """Formats the ATE, ATC, and ATT results into a Pandas DataFrame for easy viewing.

        Returns:
            DataFrame containing ATE, ATC, and ATT results
        """
        result = (self.ATE, self.ATC, self.ATT)
        if not self.validation:
            logger.info(f"The entry of bias into the ATT is {str(self.delta_t) + '%'}")

        for outcome in self.outcomes:
            res = pd.DataFrame(
                [x[outcome] + [outcome] for x in result],
                columns=[
                    "effect_size",
                    "std_err",
                    "p-val",
                    "ci_lower",
                    "ci_upper",
                    "outcome",
                ],
                index=["ATE", "ATC", "ATT"],
            )
            self.results = pd.concat([self.results, res])
        return self.results


def map_func(x: np.ndarray) -> np.ndarray:
    """Get the indices of elements in an array that are equal to the first element.

    Args:
        x:
            An input array.

    Returns:
        Array of indices where the elements match the first element of x.
    """
    return np.where(x == x[0])[0]


def f2(x: np.ndarray, y: np.ndarray) -> Any:
    """Index an array using a secondary array of indices.

    Args:
        x:
            An input array.
        y:
            Array of indices used for indexing x.

    Returns:
        Indexed element from the input array x.
    """
    return x[y]


def _get_index(base: np.ndarray, new: np.ndarray, n_neighbors: int) -> list:
    """Gets array of indexes that match a new array.

    Args:
        base:
            A numpy array serving as the reference for matching
        new:
            A numpy array that needs to be matched with the base
        n_neighbors:
            The number of neighbors to use for the matching

    Returns:
        An array of indexes containing all neighbours with minimum distance
    """
    index = faiss.IndexFlatL2(base.shape[1])
    index.add(base)
    dist, indexes = index.search(new, 20)
    if n_neighbors == 1:
        equal_dist = list(map(map_func, dist))
        indexes = [f2(i, j) for i, j in zip(indexes, equal_dist)]
    else:
        indexes = f3(indexes, dist, n_neighbors)
    return indexes


def _transform_to_np(
    treated: pd.DataFrame, untreated: pd.DataFrame, weights: dict, validation: bool
) -> Tuple[np.ndarray, np.ndarray]:
    """Transforms df to numpy and transform via Cholesky decomposition.
    If there are features that cannot be decomposed Cholesky, these features are removed.

    Args:
        treated:
            Test subset DataFrame to be transformed
        untreated:
            Control subset DataFrame to be transformed
        weights:
            Dict with weights for each feature. By default is 1
        validation:
            Flag is validation


    Returns:
        A tuple of transformed numpy arrays for treated and untreated data respectively
    """
    ## TODO: Create one more function for calculating cov
    xc = untreated.to_numpy()
    xt = treated.to_numpy()

    cov_c = np.cov(xc, rowvar=False, ddof=0)
    cov_t = np.cov(xt, rowvar=False, ddof=0)
    cov = (cov_c + cov_t) / 2

    epsilon = 1e-3

    cov = cov + np.eye(cov.shape[0]) * epsilon
    try:
        L = np.linalg.cholesky(cov)

    except:  # Вносим заглушку для удаления колонок без группировки, если таковые имеются

        treated_duplicated = list(treated.columns[treated.T.duplicated()].values)
        untreated_duplicated = list(untreated.columns[untreated.T.duplicated()].values)
        if treated_duplicated is not [] and not validation:
            logger.info(
                f"The data has the duplicate columns: {treated_duplicated} in test group"
            )
        if untreated_duplicated is not [] and not validation:
            logger.info(
                f"The data has the duplicate columns: {untreated_duplicated} in control group"
            )

        columns_duplicated = set(treated_duplicated + untreated_duplicated)
        treated = treated.drop(columns=columns_duplicated)
        untreated = untreated.drop(columns=columns_duplicated)

        xc = untreated.to_numpy()
        xt = treated.to_numpy()

        cov_c = np.cov(xc, rowvar=False, ddof=0)
        cov_t = np.cov(xt, rowvar=False, ddof=0)
        cov = (cov_c + cov_t) / 2

        epsilon = 1e-3

        cov = cov + np.eye(cov.shape[0]) * epsilon
        L = np.linalg.cholesky(cov)

    mahalanobis_transform = np.linalg.inv(L)
    if weights is not None:
        features = treated.columns
        w_list = np.array(
            [weights[col] if col in weights.keys() else 1 for col in features]
        )
        w_matrix = np.sqrt(np.diag(w_list / w_list.sum()))
        mahalanobis_transform = np.dot(w_matrix, mahalanobis_transform)
    yc = np.dot(xc, mahalanobis_transform.T)
    yt = np.dot(xt, mahalanobis_transform.T)

    return yt.copy(order="C").astype("float32"), yc.copy(order="C").astype("float32")


def calc_atx_var(
    vars_c: np.ndarray, vars_t: np.ndarray, weights_c: np.ndarray, weights_t: np.ndarray
) -> float:
    """Calculates Average Treatment Effect for the treated (ATT) variance.

    Args:
        vars_c:
            Control group variance
        vars_t:
            Treatment group variance
        weights_c:
            Control group weights
        weights_t:
            Treatment group weights

    Returns:
        The calculated ATT variance

    """
    N_c, N_t = len(vars_c), len(vars_t)
    summands_c = weights_c**2 * vars_c
    summands_t = weights_t**2 * vars_t

    return summands_t.sum() / N_t**2 + summands_c.sum() / N_c**2


def calc_atc_se(
    vars_c: np.ndarray, vars_t: np.ndarray, scaled_counts_t: np.ndarray
) -> float:
    """Calculates Average Treatment Effect for the control group (ATC) standard error.

    Args:
        vars_c:
            Control group variance
        vars_t:
            Treatment group variance
        scaled_counts_t:
            Scaled counts for treatment group

    Returns:
        The calculated ATC standard error
    """
    N_c, N_t = len(vars_c), len(vars_t)
    weights_c = np.ones(N_c)
    weights_t = (N_t / N_c) * scaled_counts_t

    var = calc_atx_var(vars_c, vars_t, weights_c, weights_t)

    return np.sqrt(var)


def conditional_covariance(xc, xt):
    """Calculates covariance according to Imbens, Rubin model."""
    cov_c = np.cov(xc, rowvar=False, ddof=0)
    cov_t = np.cov(xt, rowvar=False, ddof=0)
    cov = (cov_c + cov_t) / 2

    return cov


def calc_att_se(
    vars_c: np.ndarray, vars_t: np.ndarray, scaled_counts_c: np.ndarray
) -> float:
    """Calculates Average Treatment Effect for the treated (ATT) standard error.

    Args:
        vars_c:
            Control group variance
        vars_t:
            Treatment group variance
        scaled_counts_c:
            Scaled counts for control group

    Returns:
        The calculated ATT standard error
    """
    N_c, N_t = len(vars_c), len(vars_t)
    weights_c = (N_c / N_t) * scaled_counts_c
    weights_t = np.ones(N_t)

    var = calc_atx_var(vars_c, vars_t, weights_c, weights_t)

    return np.sqrt(var)


def calc_ate_se(
    vars_c: np.ndarray,
    vars_t: np.ndarray,
    scaled_counts_c: np.ndarray,
    scaled_counts_t: np.ndarray,
) -> float:
    """Calculates Average Treatment Effect for the control group (ATC) standard error.

    Args:
        vars_c:
            Control group variance
        vars_t:
            Treatment group variance
        scaled_counts_c:
            Scaled counts for control group
        scaled_counts_t:
            Scaled counts for treatment group

    Returns:
        The calculated ATE standard error
    """
    N_c, N_t = len(vars_c), len(vars_t)
    N = N_c + N_t
    weights_c = (N_c / N) * (1 + scaled_counts_c)
    weights_t = (N_t / N) * (1 + scaled_counts_t)

    var = calc_atx_var(vars_c, vars_t, weights_c, weights_t)

    return np.sqrt(var)


def pval_calc(z):
    """Calculates p-value of the normal cumulative distribution function based on z.

    Args:
        z:
            The z-score for which the p-value is calculated

    Returns:
        The calculated p-value rounded to 2 decimal places

    """
    return round(2 * (1 - norm.cdf(abs(z))), 2)


def scaled_counts(N: int, matches: np.ndarray, silent: bool = True) -> np.ndarray:
    """Counts the number of times each subject has appeared as a match.

    In the case of multiple matches, each subject only gets partial credit.

    Args:
        N:
            The length of original treated or control group
        matches:
            A numpy array of matched indexes from control or treated group
        silent:
            If true logger in info mode

    Returns:
        An array representing the number of times each subject has appeared as a match
    """
    s_counts = np.zeros(N)

    for matches_i in matches:
        scale = 1 / len(matches_i)
        for match in matches_i:
            s_counts[match] += scale

    if silent:
        logger.debug(
            f"Calculated the number of times each subject has appeared as a match: {len(s_counts)}"
        )
    else:
        logger.info(
            f"Calculated the number of times each subject has appeared as a match: {len(s_counts)}"
        )

    return s_counts


def bias_coefs(matches, Y_m, X_m):
    """Computes Ordinary Least Squares (OLS) coefficient in bias correction regression.

    Constructs data for regression by including (possibly multiple times) every
    observation that has appeared in the matched sample.

    Args:
        matches:
            A numpy array of matched indexes
        Y_m:
            The dependent variable values
        X_m:
            The independent variable values

    Returns:
        The calculated OLS coefficients excluding the intercept
    """
    flat_idx = np.concatenate(matches)
    N, K = len(flat_idx), X_m.shape[1]

    Y = Y_m[flat_idx]
    X = np.empty((N, K + 1))
    X[:, 0] = 1  # intercept term
    X[:, 1:] = X_m[flat_idx]

    return np.linalg.lstsq(X, Y, rcond=-1)[0][1:]  # don't need intercept coef


def bias(X, X_m, coefs):
    """Computes bias correction term.

    It is approximated by the dot product of the
    matching discrepancy (i.e., X-X_matched) and the
    coefficients from the bias correction regression.

    Args:
        X:
            The original independent variable values
        X_m:
            The matched independent variable values
        coefs:
            The coefficients from the bias correction regression

    Returns:
        The calculated bias correction terms for each observation
    """
    bias_list = [(X_j - X_i).dot(coefs) for X_i, X_j in zip(X, X_m)]

    return np.array(bias_list)


def f3(index: np.array, dist: np.array, k: int) -> list:
    """Function returns list of n matches with equal distance in case n>1.

    Args:
        index:
            Array of matched indexes
        dist:
            Array of matched distances
        k:
            k of neareast neighbors with same distance

    Returns:
        Array of indexes for k neighbors with same distance
    """
    new = []
    for i, val in enumerate(index):
        eq_dist = sorted(set(dist[i]))
        unit = []
        for d in eq_dist[:k]:
            unit.append(val[np.where(dist[i] == d)[0]])
        new.append(np.concatenate(unit))
    return new



"""
Multiple Hypothesis Testing for Best Sample Selection
=====================================================

This module outlines a statistical approach developed to select the best sample out of k samples based on a specific target metric. It is designed to adhere to predetermined Type I and Type II error probabilities, leveraging the marginal distribution of the minimum statistic. This documentation provides an overview of the problem, the proposed solution, and detailed descriptions of the main functions implemented.

Problem Statement:
------------------
In multiple hypothesis testing, our aim is to identify the superior sample from a set of k samples, each evaluated according to a certain target metric. We denote by H_i the hypothesis that the i-th sample is the best among all, for i = 1, ..., k, and by H_0 the null hypothesis that there is no single best sample among the group.

Methodology:
------------
The method is based on the limit distribution theory, allowing us to maintain the predefined probabilities of committing Type I and Type II errors. The core approach involves calculating the minimum sample size required for a reliable comparison and applying a criterion to test the hypotheses, both of which rely on the quantile function of the marginal distribution of the minimum statistic.

Functions Overview:
-------------------
1. Quantile of Marginal Distribution: Calculates the quantile of the marginal distribution of the minimum of a set of random variables. Essential for determining the critical values used in the main hypothesis testing and sample size calculation functions.

2. Minimum Sample Size Calculation: Determines the smallest number of observations required in each sample to reliably detect a specified effect size with a given power and significance level.

3. Hypothesis Testing Criterion: Tests whether any of the samples is significantly better than the others based on their target metric distributions.

Implementation:
---------------
Implemented in Python, utilizing numpy for numerical computations and scipy.stats for statistical functions. Detailed function descriptions, including input parameters and return values, are documented within the function docstrings in this file.

Mathematical Formulations:
--------------------------
The quantile calculation for the marginal distribution of the minimum statistic is based on the following principle:

Q(p; F_1, ..., F_k) = min { q_1(p), ..., q_k(p) }

where Q(p; F_1, ..., F_k) represents the p-th quantile of the marginal distribution of the minimum of k random variables with distribution functions F_1, ..., F_k, and q_i(p) denotes the p-th quantile of the i-th distribution.

Conclusion:
-----------
This module provides a robust solution for statistical comparison across multiple samples, ensuring adherence to specified error probabilities through the application of limit distribution theory.

"""

from typing import Optional, List, Union

import numpy as np
from scipy.stats import norm


def quantile_of_marginal_distribution(
    num_samples: int,
    quantile_level: float,
    variances: Optional[Union[List, float]] = None,
    iteration_size: int = 20000,
    random_state: int = 42,
):
    """Calculate the quantile(s) of the marginal distribution for minimum t-values across multiple comparisons.

    This function generates random samples from a normal distribution and computes t-values for comparisons either
    with equal variances (if variances are not provided) or with specified unequal variances (if variances are provided).
    It then determines the quantile of interest for the distribution of the minimum t-values from these comparisons.

    Args:
        num_samples: The number of samples/groups to compare.
        quantile_level: The quantile level to compute for the marginal distribution (e.g., 0.95 for the 95th percentile).
        variances: A list of variances for each sample/group. If None, equal variances are assumed.
        iteration_size: The number of iterations/random samples to generate for the simulation.
        random_state: Optional parameter for data randomisation.

    Returns:
       The quantile of interest for the marginal distribution of the minimum t-values. Returns
       a single float if variances are assumed equal (or not provided) or a list of floats
       with quantiles for each sample if variances are provided and unequal.
    """
    np.random.seed(random_state)
    total = norm.rvs(size=[iteration_size, num_samples])

    if variances is None:
        j = 0
        t_values = [
            min(
                [
                    (total[l][j] - total[l][i]) / np.sqrt(2)
                    for i in range(num_samples)
                    if i != j
                ]
            )
            for l in range(iteration_size)
        ]
        return np.quantile(t_values, quantile_level)

    quantiles = []
    for j in range(num_samples):
        t_values = [
            min(
                [
                    total[l][j] / np.sqrt(1 + variances[i] / variances[j])
                    - total[l][i] / np.sqrt(1 + variances[j] / variances[i])
                    for i in range(num_samples)
                    if i != j
                ]
            )
            for l in range(iteration_size)
        ]
        quantiles += [np.quantile(t_values, quantile_level)]
    return quantiles


def test_on_marginal_distribution(
    samples: List[np.ndarray],
    significance_level: float = 0.05,
    equal_variance: bool = True,
    quantiles: Optional[Union[float, List[float]]] = None,
) -> int:
    """Performs a test on the marginal distribution of minimum t-values across multiple samples/groups.

    This function calculates the means and variances for each sample/group, determines the quantile of interest for
    the marginal distribution of the minimum t-values, and identifies if any sample's minimum t-value exceeds this
    quantile. It is used to assess whether there is a statistically significant difference between any of the samples.

    Args:
        samples: A list of arrays, where each array represents the data of a sample/group.
        significance_level: The significance level for the test (default is 0.05).
        equal_variance: A boolean indicating if the samples are assumed to have equal variance (default is True).
        quantiles: Pre-computed quantiles of the marginal distribution. If None, they will be computed.

    Returns:
        The index of the first sample that significantly differs from others, or 0 if none are found.
    """

    num_samples, sample_size = len(samples), len(samples[0])

    means = [np.mean(sample) for sample in samples]
    variances = [np.var(sample) * sample_size / (sample_size - 1) for sample in samples]
    variances_q = variances if equal_variance else None

    if quantiles is None or isinstance(quantiles, float):
        quantiles = quantile_of_marginal_distribution(
            num_samples=num_samples,
            quantile_level=1 - significance_level / num_samples,
            variances=variances_q,
        )
    for j in range(num_samples):
        total_j = np.inf
        for i in range(num_samples):
            if i != j:
                total = (
                    np.sqrt(sample_size)
                    * (means[j] - means[i])
                    / np.sqrt(variances[j] + variances[i])
                )
                total_j = total if total < total_j else total_j
        if total_j > quantiles[j]:
            return j + 1
    return 0


def min_sample_size(
    number_of_samples: int,
    minimum_detectable_effect: float,
    variances: Union[List[float], float],
    significance_level: float = 0.05,
    power_level: float = 0.2,
    equal_variance: bool = True,
    quantile_1: Optional[Union[float, List[float]]] = None,
    quantile_2: Optional[float] = None,
    initial_estimate: Optional[int] = None,
    random_state: int = 42,
) -> int:
    """
    Calculates the minimum sample size required to detect a given effect with specified power and significance level.

    The function computes the minimum sample size for either equal or unequal variances across samples. It can use
    pre-computed quantiles of the marginal distribution to speed up calculations.

    Args:
        number_of_samples: Number of samples or groups to compare.
        minimum_detectable_effect: The minimum effect size that the test is designed to detect.
        variances: Variances of the samples. Can be a single value if equal_variance is True, or a list of variances.
        significance_level: The significance level for hypothesis testing (alpha).
        power_level: Desired power of the test (1 - beta).
        equal_variance: Boolean flag indicating whether to assume equal variances across samples.
        quantile_1: Optional pre-computed quantile for the significance level. Calculated if None.
        quantile_2: Optional pre-computed quantile for the power level. Calculated if None.
        initial_estimate: Optional initial estimate for the sample size to speed up calculations.
        random_state: Optional parameter for data randomisation.

    Returns:
        The minimum sample size required per sample/group.
    """
    if equal_variance:

        if quantile_1 is None:
            quantile_1 = quantile_of_marginal_distribution(
                num_samples=number_of_samples,
                quantile_level=1 - significance_level / number_of_samples,
            )  # quantile of the marginal distribution 1-alpha/k

        if quantile_2 is None:
            quantile_2 = quantile_of_marginal_distribution(
                num_samples=number_of_samples, quantile_level=power_level
            )  # quantile of the marginal distribution beta

        return (
            int(
                2
                * variances
                * ((quantile_1 - quantile_2) / minimum_detectable_effect) ** 2
            )
            + 1
        )
    else:
        iteration_size = 3000  # number of iterations
        if quantile_1 is None:
            quantile_1 = quantile_of_marginal_distribution(
                num_samples=number_of_samples,
                quantile_level=1 - significance_level / number_of_samples,
                variances=variances,
            )  # set of quantiles of the marginal distribution
        sample_sizes = []  # for sample sizes
        for j in range(number_of_samples):
            sample_size = initial_estimate or 0
            current_power = 0  # power
            while current_power < 1 - power_level:
                sample_size += 100
                current_power = 0
                total_samples = norm.rvs(
                    size=[iteration_size, number_of_samples], random_state=random_state
                )
                for sample in total_samples:
                    min_t_value = np.inf
                    for i in range(number_of_samples):
                        if i != j:
                            t_value = (
                                sample[j] / np.sqrt(1 + variances[i] / variances[j])
                                - sample[i] / np.sqrt(1 + variances[j] / variances[i])
                                + minimum_detectable_effect
                                * np.sqrt(sample_size / (variances[j] + variances[i]))
                            )
                            min_t_value = min(min_t_value, t_value)
                    if min_t_value > quantile_1[j]:
                        current_power += 1
                current_power /= iteration_size
            sample_sizes.append(sample_size)
        return np.max(sample_sizes)



"""Module for AA tests"""

import warnings
from itertools import combinations
from pathlib import Path
from typing import Iterable, Union, Optional, Dict, Any, Tuple, List

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from scipy.stats import ttest_ind, ks_2samp, norm
from sklearn.utils import shuffle
from statsmodels.stats.power import TTestIndPower

try:
    from tqdm.auto import tqdm
except:
    try:
        from tqdm import tqdm
    except:
        raise Exception("Can't import tqdm")


class AATest:
    """
    A class for conducting AA testing (random split testing) to assess the
    statistical uniform of two samples.

    AA testing is used to validate that the splitting mechanism of an A/B test
    is unbiased and random. This class supports various statistical methods to
    evaluate the equivalence of two randomly split samples in terms of various
    metrics.

    Attributes:
        target_fields (Union[Iterable[str], str]): Target column names to analyze. This fields should be numeric.
        group_cols (Union[str, Iterable[str]]): Column names used for grouping. This fields should be categorical. It's a field for stratification. Stratification - the way to divide groups with equal number of categories in each of them.
        info_cols (Union[str, Iterable[str]]): Column names for additional information.
        quant_field (str): Name of the column for quantization. This fields should be categorical. A quantum is a category that passes entirely into one of the groups.
        mode (str): Mode of the AA test. Options are 'simple' and 'balanced'. 'simple' - naively splits groups in half. 'balanced' - separation with quantum balancing at each step (only used if a quantization field is specified.
        alpha (float): Level of significance for statistical tests.

    Methods:
        columns_labeling(data):
            Classifies columns in the input DataFrame as target or group columns.
        __simple_mode(data, random_state, test_size):
            Internal method to create a simple random split of the data.
        split(data, random_state, test_size):
            Splits the dataset into test and control groups.
        merge_groups(control_group, test_group):
            Merges test and control groups in one DataFrame and creates column "group".
        _postprep_data(data, spit_indexes):
            Combines the index markup obtained at the split step.
        calc_ab_delta(a_mean, b_mean, mode):
            Calculates the difference between averages of two samples.
        sampling_metrics(data, random_state, test_size):
            Computes various metrics for a single random split of the data.
        calc_uniform_tests(data, test_size, iterations, file_name, experiment_write_mode, split_write_mode, write_step, pbar):
            Runs multiple iterations of AA tests to find a uniform distribution.
        features_p_value_distribution(experiment_results, figsize, bin_step):
            Plots the distribution of p-values for each feature.
        aa_score(experiment_results):
            Computes the average score for passed tests in AA testing.
        uniform_tests_interpretation(experiment_results):
            Analyzes and plots the results of uniform tests.
        num_feature_uniform_analysis(control_data, test_data, plot_set):
            Analyzes and plots numerical feature distributions in control and test data.
        cat_feature_uniform_analysis(control_data, test_data):
            Analyzes and plots categorical feature distributions in control and test data.
        experiment_result_transform(experiment):
            Transforms the result of an experiment into a readable format.
        split_analysis(splited_data):
            Analyzes split data for both target and group columns.
        get_resume(aa_score, best_experiment_stat):
            Formats the final results of AA testing for clarity.
        process(data, optimize_groups, iterations, show_plots, test_size, pbar):
            Main method to perform the complete AA test process, including optimization, testing, and result presentation.

    Example:
        >>> aa_test = AATest(target_fields=["metric1", "metric2"], group_cols=["group"], info_cols=["info1", "info2"])
        >>> results = aa_test.process(data, optimize_groups=True, iterations=1000, show_plots=True)
    """

    def __init__(
        self,
        target_fields: Union[Iterable[str], str] = None,
        group_cols: Union[str, Iterable[str]] = None,
        info_cols: Union[str, Iterable[str]] = None,
        quant_field: str = None,
        mode: str = "simple",
        alpha: float = 0.05,
    ):
        """Initialize the AATest class.

        Args:
            target_fields:
                List or str with target columns. This fields should be numeric.
            group_cols:
                List or str with columns for grouping. This fields should be categorical. It's a field for stratification. Stratification - the way to divide groups with equal number of categories in each of them.
            info_cols:
                List or str with informational columns
            quant_field:
                String with name of column for quantization. This fields should be categorical. A quantum is a category that passes entirely into one of the groups.
            mode:
                Mode of the AA-test
                Available modes:
                    * simple - naively splits groups in half
                    * balanced - separation with quantum balancing at each step (only used if a quantization field is specified)
            alpha:
                Level of significance
        """
        self.target_fields = (
            [target_fields] if isinstance(target_fields, str) else target_fields
        )
        self.group_cols = (
            [group_cols] if isinstance(group_cols, str) else group_cols
        ) or []
        self.info_cols = [info_cols] if isinstance(info_cols, str) else info_cols
        self.quant_field = quant_field
        self.mode = mode
        self.alpha = alpha

    def columns_labeling(self, data: pd.DataFrame) -> Dict[str, List[str]]:
        """
        Label columns as target columns and group columns.

        Args:
            data:
                Input dataframe

        Returns:
            Dictionary with list of target columns and group columns

        """
        return {
            "target_field": list(
                data.select_dtypes(include="number").columns.drop(
                    self.info_cols, errors="ignore"
                )
            ),
            "group_col": list(
                data.select_dtypes(include="object").columns.drop(
                    self.info_cols, errors="ignore"
                )
            ),
        }

    def __simple_mode(
        self, data: pd.DataFrame, random_state: int = None, test_size: float = 0.5
    ) -> Dict:
        """Separates data on A and B samples within simple mode.
        Separation performed to divide groups of equal sizes - equal amount of records
        or equal amount of groups in each sample.

        Args:
            data:
                Input data
            random_state:
                Seed of random

        Returns:
            Test and control samples of indexes dictionary
        """
        result = {"test_indexes": [], "control_indexes": []}

        if self.quant_field:
            random_ids = shuffle(
                data[self.quant_field].unique(), random_state=random_state
            )
            edge = int(len(random_ids) * test_size)
            result["test_indexes"] = list(
                data[data[self.quant_field].isin(random_ids[:edge])].index
            )
            result["control_indexes"] = list(
                data[data[self.quant_field].isin(random_ids[edge:])].index
            )

        else:
            addition_indexes = list(shuffle(data.index, random_state=random_state))
            edge = int(len(addition_indexes) * test_size)
            result["test_indexes"] = addition_indexes[:edge]
            result["control_indexes"] = addition_indexes[edge:]

        return result

    def split(
        self, data: pd.DataFrame, random_state: int = None, test_size: float = 0.5
    ) -> Dict:
        """Divides sample on two groups.

        Args:
            data:
                Input data
            random_state:
                Seed of random - one integer to fix split
            test_size:
                Proportion of the test group

        Returns:
            Dict of indexes with division on test and control group
        """
        result = {"test_indexes": [], "control_indexes": []}

        if self.group_cols:
            groups = data.groupby(self.group_cols)
            for _, gd in groups:
                if self.mode not in ("balanced", "simple"):
                    warnings.warn(
                        f"The mode '{self.mode}' is not supported for group division. Implemented mode 'simple'."
                    )
                    self.mode = "simple"

                if self.mode == "simple":
                    t_result = self.__simple_mode(gd, random_state, test_size)
                    result["test_indexes"] += t_result["test_indexes"]
                    result["control_indexes"] += t_result["control_indexes"]

                elif self.mode == "balanced":
                    if self.quant_field:
                        random_ids = shuffle(
                            gd[self.quant_field].unique(), random_state=random_state
                        )
                        addition_indexes = list(
                            gd[gd[self.quant_field].isin(random_ids)].index
                        )
                    else:
                        addition_indexes = list(
                            shuffle(gd.index, random_state=random_state)
                        )

                    if len(result["control_indexes"]) > len(result["test_indexes"]):
                        result["test_indexes"] += addition_indexes
                    else:
                        result["control_indexes"] += addition_indexes

        else:
            if self.mode != "simple":
                warnings.warn(
                    f"The mode '{self.mode}' is not supported for regular division. "
                    f"Implemented mode 'simple'."
                )

            t_result = self.__simple_mode(data, random_state, test_size)
            result["test_indexes"] = t_result["test_indexes"]
            result["control_indexes"] = t_result["control_indexes"]

        result["test_indexes"] = list(set(result["test_indexes"]))
        result["control_indexes"] = list(set(result["control_indexes"]))

        return result

    @staticmethod
    def merge_groups(
        control_group: Union[Iterable[pd.DataFrame], pd.DataFrame],
        test_group: Union[Iterable[pd.DataFrame], pd.DataFrame],
    ) -> pd.DataFrame:
        """Merges test and control groups in one DataFrame and creates column "group".
        Column "group" contains of "test" and "control" values.

        Args:
            control_group:
                Data of control group
            test_group:
                Data of target group
        Returns:
            merged_data:
                Contacted DataFrame
        """
        control_group.loc[:, "group"] = "control"
        test_group.loc[:, "group"] = "test"

        return pd.concat([test_group, control_group], ignore_index=True)

    def _postprep_data(self, data, spit_indexes: Dict = None) -> pd.DataFrame:
        """Prepares data to show user.
        Adds info_cols and decode binary variables.

        Args:
            data:
                Input data
            spit_indexes:
                Dict of indexes with separation on test and control group

        Returns:
            Separated initial data with column "group"
        """
        test = data.loc[spit_indexes["test_indexes"]]
        control = data.loc[spit_indexes["control_indexes"]]
        data = self.merge_groups(control, test)

        return data

    @staticmethod
    def calc_ab_delta(a_mean: float, b_mean: float, mode: str = "percentile") -> float:
        """Calculates target delta between A and B groups.

        Args:
            a_mean:
                Average of target in one group
            b_mean:
                Average of target in another group
            mode:
                Type of expected result:
                    * 'percentile' - percentage exceeding the average in group A compared to group B
                    * 'absolute' - absolute value of difference between B and A group
                    * 'relative' - percent in format of number (absolute) exceeding the average in group A compared to group B

        Returns:
            Delta between groups as percent or absolute value
        """
        if mode == "percentile":
            return (1 - a_mean / b_mean) * 100
        if mode == "absolute":
            return b_mean - a_mean
        if mode == "relative":
            return 1 - a_mean / b_mean

    def sampling_metrics(
        self, data: pd.DataFrame, random_state: int = None, test_size: float = 0.5
    ) -> Dict:
        """Calculates metrics of one sampling.

        Args:
            data:
                Input data
            random_state:
                Random seeds for searching
            test_size:
                Proportion of the test group

        Returns:
            Dict of
                1) metrics dataframe (stat tests) and
                2) dict of random state with test_control dataframe
        """
        scores = []
        t_result = {"random_state": random_state}

        split = self.split(data, random_state, test_size)

        a = data.loc[split["control_indexes"]]
        b = data.loc[split["test_indexes"]]

        data_from_sampling_dict = {random_state: self._postprep_data(data, split)}
        for tf in self.target_fields:
            ta = a[tf]
            tb = b[tf]

            t_result[f"{tf} a mean"] = ta.mean()
            t_result[f"{tf} b mean"] = tb.mean()
            t_result[f"{tf} ab delta"] = self.calc_ab_delta(
                t_result[f"{tf} a mean"], t_result[f"{tf} b mean"], "absolute"
            )
            t_result[f"{tf} ab delta %"] = self.calc_ab_delta(
                t_result[f"{tf} a mean"], t_result[f"{tf} b mean"], "percentile"
            )
            t_result[f"{tf} t-test p-value"] = ttest_ind(
                ta, tb, nan_policy="omit"
            ).pvalue
            t_result[f"{tf} ks-test p-value"] = ks_2samp(ta, tb).pvalue
            t_result[f"{tf} t-test passed"] = (
                t_result[f"{tf} t-test p-value"] < self.alpha
            )
            t_result[f"{tf} ks-test passed"] = (
                t_result[f"{tf} ks-test p-value"] < self.alpha
            )
            scores.append(
                (
                    t_result[f"{tf} t-test p-value"]
                    + 2 * t_result[f"{tf} ks-test p-value"]
                )
                / 3
            )

        t_result["control %"] = len(a) / len(data) * 100
        t_result["test %"] = len(b) / len(data) * 100
        t_result["control size"] = len(a)
        t_result["test size"] = len(b)
        t_result["t-test mean p-value"] = np.mean(
            [p_value for key, p_value in t_result.items() if "t-test p-value" in key]
        )
        t_result["ks-test mean p-value"] = np.mean(
            [p_value for key, p_value in t_result.items() if "ks-test p-value" in key]
        )
        t_result["t-test passed %"] = np.mean(
            [passed * 100 for key, passed in t_result.items() if "t-test passed" in key]
        )
        t_result["ks-test passed %"] = np.mean(
            [
                passed * 100
                for key, passed in t_result.items()
                if "ks-test passed" in key
            ]
        )
        t_result["mean_tests_score"] = np.mean(scores)
        return {"metrics": t_result, "data_from_experiment": data_from_sampling_dict}

    def calc_uniform_tests(
        self,
        data: pd.DataFrame,
        test_size: float = 0.5,
        iterations: int = 2000,
        file_name: Union[Path, str] = None,
        experiment_write_mode: str = "full",
        split_write_mode: str = "full",
        write_step: int = None,
        pbar: bool = True,
        **kwargs,
    ) -> Optional[Tuple[pd.DataFrame, Dict[Any, Dict]]]:
        """Performs multiple separation experiments for different random states.

        Args:
            data:
                Input data
            iterations:
                Number of iterations to search uniform sampling to searching
            test_size:
                Proportion of the test group
            file_name:
                Name of file to save results (if None - no results will be saved, func returns result)
            experiment_write_mode:
                Mode to write experiment results:
                    'full' - save all experiments
                    'all' - save experiments that passed all statistical tests
                    'any' - save experiments that passed any statistical test
            split_write_mode:
                Mode to write split results:
                    'full' - save all experiments
                    'all' - save experiments that passed all statistical tests
                    'any' - save experiments that passed any statistical test
            write_step:
                Step to write experiments to file
            pbar:
                Flag to show progress bar

        Returns:
                If no saving (no file_name, no write mode and no write_step) returns dataframe
                else None and saves file to csv
        """
        random_states = range(iterations)
        results = []
        data_from_sampling = {}

        if experiment_write_mode not in ("full", "all", "any"):
            warnings.warn(
                f"Write mode '{experiment_write_mode}' is not supported. Mode 'full' will be used"
            )
            experiment_write_mode = "full"
        if split_write_mode not in ("full", "all", "any"):
            warnings.warn(
                f"Write mode '{split_write_mode}' is not supported. Mode 'full' will be used"
            )
            split_write_mode = "full"

        for i, rs in tqdm(
            enumerate(random_states), total=len(random_states), disable=not pbar
        ):
            res = self.sampling_metrics(data, random_state=rs, test_size=test_size)

            # write to file
            passed = []
            for tf in self.target_fields:
                passed += [
                    not res["metrics"][f"{tf} t-test passed"],
                    not res["metrics"][f"{tf} ks-test passed"],
                ]

            if all(passed):
                if experiment_write_mode == "all":
                    results.append(res["metrics"])
                if split_write_mode == "all":
                    data_from_sampling.update(res["data_from_experiment"])
            if any(passed):
                if experiment_write_mode == "any":
                    results.append(res["metrics"])
                if split_write_mode == "any":
                    data_from_sampling.update(res["data_from_experiment"])
            if experiment_write_mode == "full":
                results.append(res["metrics"])
            if split_write_mode == "full":
                data_from_sampling.update(res["data_from_experiment"])

            if file_name and write_step:
                if i == write_step:
                    pd.DataFrame(results).to_csv(file_name, index=False)
                elif i % write_step == 0:
                    pd.DataFrame(results).to_csv(
                        file_name, index=False, header=False, mode="a"
                    )
                    results = []

        results = pd.DataFrame(results)
        if file_name and write_step:
            results.to_csv(file_name, index=False, header=False, mode="a")
        elif file_name:
            results.to_csv(file_name, index=False)
            return results, data_from_sampling
        else:
            return results, data_from_sampling

    def features_p_value_distribution(
        self, experiment_results: pd.DataFrame, figsize=None, bin_step=0.05
    ):
        """Process plots of features' p-value distribution.

        Args:
            experiment_results:
                Results of experiments
            figsize:
                Size of figure for plot
            bin_step:
                Step for bins in X axis
        """
        feature_num = len(self.target_fields)
        figsize = figsize or (15, 7 * feature_num)
        bin_step = bin_step or self.alpha
        bins = np.arange(0, 1 + bin_step, bin_step)
        figure, axs = plt.subplots(nrows=feature_num, ncols=2, figsize=figsize)
        for i in range(feature_num):
            sns.histplot(
                data=experiment_results,
                x=f"{self.target_fields[i]} t-test p-value",
                ax=axs[i, 0],
                bins=bins,
                stat="percent",
                shrink=0.8,
            )
            sns.histplot(
                data=experiment_results,
                x=f"{self.target_fields[i]} ks-test p-value",
                ax=axs[i, 1],
                bins=bins,
                stat="percent",
                shrink=0.8,
            )

            axs[i, 0].set_title(
                f"{self.target_fields[i]} t-test p-value\npassed score: {experiment_results[f'{self.target_fields[i]} t-test passed'].mean():.3f}"
            )
            axs[i, 1].set_title(
                f"{self.target_fields[i]} ks-test p-value\npassed score: {experiment_results[f'{self.target_fields[i]} ks-test passed'].mean():.3f}"
            )
        plt.show()

    def aa_score(self, experiment_results: pd.DataFrame) -> pd.DataFrame:
        """Estimates mean passed score for t-test and ks-test in AA-test.

        Args:
            experiment_results:
                Results of the experiment

        Returns:
            Pandas dataframe containing the results of the AA-test
        """
        result = pd.DataFrame(
            {
                f: {
                    "t-test passed score": experiment_results[
                        f"{f} t-test passed"
                    ].mean(),
                    "ks-test passed score": experiment_results[
                        f"{f} ks-test passed"
                    ].mean(),
                }
                for f in self.target_fields
            }
        ).T

        result["t-test aa passed"] = result["t-test passed score"].apply(
            lambda x: 0.8 * self.alpha <= x <= 1.2 * self.alpha
        )
        result["ks-test aa passed"] = result["ks-test passed score"].apply(
            lambda x: 0.8 * self.alpha <= x <= 1.2 * self.alpha
        )
        result.loc["mean"] = result.mean()

        return result

    def uniform_tests_interpretation(
        self, experiment_results: pd.DataFrame, **kwargs
    ) -> pd.DataFrame:
        """Process plotting of p-value distribution and results of AA-test.

        Args:
            experiment_results:
                Results of experiments
            **kwargs:
                Some extra keyword arguments:
                    * figsize: Size of figure for plot
                    * bin_step: Step for bins in X axis

        Returns:
            Pandas dataframe containing the results of the AA-test
        """
        self.features_p_value_distribution(
            experiment_results,
            figsize=kwargs.get("figsize"),
            bin_step=kwargs.get("bin_step"),
        )
        return self.aa_score(experiment_results)

    def num_feature_uniform_analysis(
        self,
        control_data: pd.Series,
        test_data: pd.Series,
        plot_set: Tuple = ("hist", "cumulative", "percentile"),
        **kwargs,
    ):
        """Show plots of distribution in groups with uniform tests.

        Args:
            control_data:
                Data from control group
            test_data:
                Data from test group
            plot_set:
                Type of plot
                Available types:
                    * hist
                    * cumulative
                    * percentile
            **kwargs:
                Some extra keyword arguments:
                    * figsize: Size of figure for plot
                    * bins: Number of bins in X axis
                    * alpha: Transparency of histograms
        """
        if not plot_set:
            return

        figsize = kwargs.get("figsize", (25, 20))
        figure, axs = plt.subplots(
            nrows=len(plot_set),
            ncols=1,
            figsize=figsize,
            facecolor="honeydew",
            edgecolor="black",
        )
        ax_count = 0

        bins = np.arange(
            min(control_data.min(), test_data.min()),
            max(control_data.max(), test_data.max()),
            (
                max(control_data.max(), test_data.max())
                - min(control_data.min(), test_data.min())
            )
            / kwargs.get("bins", 100),
        )

        if "hist" in plot_set:
            sns.histplot(
                data=control_data,
                ax=axs[ax_count],
                bins=bins,
                stat="percent",
                element="poly",
                alpha=kwargs.get("alpha", 0.3),
                color="blue",
            )
            sns.histplot(
                data=test_data,
                ax=axs[ax_count],
                bins=bins,
                stat="percent",
                element="poly",
                alpha=kwargs.get("alpha", 0.3),
                color="red",
            )
            axs[ax_count].grid(True)
            axs[ax_count].legend(["control", "test"])
            axs[ax_count].set_title("Histogram")
            ax_count += 1

        if "cumulative" in plot_set:
            sns.histplot(
                data=control_data,
                ax=axs[ax_count],
                bins=bins,
                stat="percent",
                element="poly",
                cumulative=True,
                alpha=kwargs.get("alpha", 0.3),
                color="blue",
            )
            sns.histplot(
                data=test_data,
                ax=axs[ax_count],
                bins=bins,
                stat="percent",
                element="poly",
                cumulative=True,
                alpha=kwargs.get("alpha", 0.3),
                color="red",
            )
            axs[ax_count].legend(["control", "test"])
            axs[ax_count].set_title("Cumulative distribution")
            ax_count += 1

        if "percentile" in plot_set:
            axs[ax_count].fill_between(
                range(101),
                [control_data.quantile(q) for q in np.arange(0, 1.01, 0.01)],
                color="blue",
                alpha=kwargs.get("alpha", 0.3),
            )
            axs[ax_count].fill_between(
                range(101),
                [test_data.quantile(q) for q in np.arange(0, 1.01, 0.01)],
                color="red",
                alpha=kwargs.get("alpha", 0.3),
            )
            axs[ax_count].legend(["control", "test"])
            axs[ax_count].set_xticks(np.arange(0, 101))
            axs[ax_count].set_xticklabels(np.arange(0, 101), rotation=45)
            axs[ax_count].set_title("Percentile distribution")

        fig_title = f"""{control_data.name}

            t-test p-value: {ttest_ind(control_data, test_data, nan_policy='omit').pvalue:.3f}
            ks-test p-value: {ks_2samp(control_data, test_data).pvalue:.3f}"""
        figure.suptitle(fig_title, fontsize=kwargs.get("title_size", 20))
        plt.show()

    def cat_feature_uniform_analysis(
        self, control_data: pd.Series, test_data: pd.Series, **kwargs
    ):
        """Show plots of distribution in groups.

        Args:
            control_data:
                Data from control group
            test_data:
                Data from test group
            **kwargs:
                Some extra keyword arguments:
                    * figsize: Size of figure for plot
                    * alpha: Transparency of histograms
        """
        s_control_data = control_data.astype("str")
        s_test_data = test_data.astype("str")

        figsize = kwargs.get("figsize", (15, 10))
        figure, ax = plt.subplots(
            nrows=1, ncols=1, figsize=figsize, facecolor="honeydew", edgecolor="black"
        )

        control_counts = s_control_data.value_counts(normalize=True) * 100
        test_counts = s_test_data.value_counts(normalize=True) * 100

        ax.fill_between(
            control_counts.index,
            control_counts.values,
            color="blue",
            alpha=kwargs.get("alpha", 0.3),
            label="control",
        )
        ax.fill_between(
            test_counts.index,
            test_counts[
                [i for i in test_counts.index if i in control_counts.index]
            ].values,
            color="red",
            alpha=kwargs.get("alpha", 0.3),
            label="test",
        )

        ax.legend()
        ax.tick_params(axis="x", rotation=90)
        figure.suptitle(f"{control_data.name}", fontsize=kwargs.get("title_size", 20))
        plt.show()

    def __get_test_and_control_series(
        self,
        test_group: pd.Series = None,
        control_group: pd.Series = None,
        data: pd.DataFrame = None,
        target_field: str = None,
    ):
        if test_group is None and control_group is None:
            if data is None or target_field is None:
                raise ValueError(
                    "test_group and control_group cannot be None if data and target_field are None"
                )
            splited_data = self.split_splited_data(data)
            test_group = splited_data["test"][target_field]
            control_group = splited_data["control"][target_field]
        return {"test_group": test_group, "control_group": control_group}

    def calc_mde(
        self,
        test_group: pd.Series = None,
        control_group: pd.Series = None,
        data: pd.DataFrame = None,
        target_field: str = None,
        reliability: float = 0.95,
        power: float = 0.8,
    ) -> float:
        """Calculates the minimum detectable effect (MDE) for a given test and control groups.

        Args:
            test_group:
                The test group as a pandas Series
            control_group:
                The control group as a pandas Series
            data:
                The input data as a pandas DataFrame used if test_group and control_group are None
            target_field:
                The target field used if given data is a DataFrame
            reliability:
                The reliability of the test
            power:
                The power of the test

        Returns:
            The minimum detectable effect
        """

        m = norm.ppf(1 - (1 - reliability) / 2) + norm.ppf(power)

        groups = self.__get_test_and_control_series(
            test_group=test_group,
            control_group=control_group,
            data=data,
            target_field=target_field,
        )
        test_group = groups["test_group"]
        control_group = groups["control_group"]

        n_test, n_control = len(test_group), len(control_group)
        proportion = n_test / (n_test + n_control)
        p = np.sqrt(1 / (proportion * (1 - proportion)))

        var_test, var_control = np.var(test_group, ddof=1), np.var(
            control_group, ddof=1
        )
        s = np.sqrt(var_test / n_test + var_control / n_control)

        return p * m * s

    def calc_sample_size(
        self,
        test_group: pd.Series = None,
        control_group: pd.Series = None,
        data: pd.DataFrame = None,
        target_field: str = None,
        mde=None,
        significance: float = 0.05,
        power: float = 0.8,
    ) -> float:
        """Calculates sample size of dataframe depends on mde and power.

        Args:
            test_group:
                The test group as a pandas Series
            control_group:
                The control group as a pandas Series
            data:
                The input data as a pandas DataFrame used if test_group and control_group are None
            target_field:
                The target field used if given data is a DataFrame
            mde:
                Minimum detectable effect. If None, it will be calculated. If it is a tuple, it will be used as relative mde
            significance:
                Level of significance (alpha)
            power:
                Power of criterion

        Returns:
            Sample_size of dataframe
        """
        if isinstance(mde, Iterable):
            z_alpha = norm.ppf((2 - significance) / 2)
            z_beta = norm.ppf(power)

            p1 = mde[0]
            p2 = mde[1]

            return (
                (z_alpha + z_beta) ** 2
                * (p1 * (1 - p1) + p2 * (1 - p2))
                / (p1 - p2) ** 2
            )
        else:
            groups = self.__get_test_and_control_series(
                test_group=test_group,
                control_group=control_group,
                data=data,
                target_field=target_field,
            )
            control_group = groups["control_group"]
            test_group = groups["test_group"]

            mde = mde or self.calc_mde(
                test_group=test_group,
                control_group=control_group,
                reliability=1 - significance,
                power=power,
            )

            control_std = control_group.std()
            test_std = test_group.std()

            test_proportion = len(test_group) / (len(test_group) + len(control_group))
            control_proportion = 1 - test_proportion

            d = ((norm.ppf(1 - significance / 2) + norm.ppf(power)) / mde) ** 2
            s = test_std**2 / test_proportion + control_std**2 / control_proportion
            return d * s

    def calc_imbalanced_sample_size(
        self,
        target_data: pd.Series,
        expected_mean: float,
        proportion: float = 0.5,
        power: float = 0.8,
    ) -> Tuple:
        """Calculates imbalanced sample size for control and test group.

        Args:
            target_data:
                The control group as a pandas Series
            expected_mean:
                Expected conversion of test group
            proportion:
                Proportion of control group
            power:
                Power of criterion

        Returns:
            Tuple with size for control and test group
        """
        target_mean = target_data.mean()

        if target_mean == expected_mean:
            raise ValueError("Current conversion and expected conversion are equal!")

        proportion = (1 - proportion) / proportion
        z_alpha = norm.ppf(1 - self.alpha / 2)
        z_power = norm.ppf(power)
        if target_data.nunique() == 2:
            h_cohen = 2 * np.arcsin(target_mean**0.5) - 2 * np.arcsin(
                expected_mean**0.5
            )
            control_size = (
                (1 + proportion) / proportion * ((z_alpha + z_power) / h_cohen) ** 2
            )
        else:
            mde = abs(expected_mean - target_mean)
            control_size = (
                (1 + proportion)
                / proportion
                * (target_data.std() ** 2)
                * (z_alpha + z_power) ** 2
                / mde**2
            )
        test_size = proportion * control_size

        return np.int32(np.ceil(control_size)), np.int32(np.ceil(test_size))

    @staticmethod
    def calc_power(
        effect_size: float,
        control_size: float,
        test_size: float,
        significance: float = 0.05,
    ) -> float:
        """Statistical power calculations for t-test for two independent sample and known significance.

        Args:
            effect_size:
                Size of the effect
            control_size:
                Size of control group
            test_size:
                Size of test group
            significance:
                Level of significance (alpha)

        Returns:
            Statistical power
        """
        analysis = TTestIndPower()
        ratio = test_size / control_size
        return analysis.power(
            effect_size=effect_size,
            nobs1=test_size,
            ratio=ratio,
            alpha=significance,
        )

    def experiment_result_transform(self, experiment: pd.Series):
        """
        Transform experiments results into readable view.

        Args:
            experiment:
                Results of experiments

        Returns:
            DataFrame with results of the experiment and statistics from best split
        """
        targets_dict = {}
        for tf in self.target_fields:
            targets_dict[tf] = {}
            for i in experiment.index:
                if i.startswith(f"{tf} "):
                    targets_dict[tf][i[len(tf) + 1 :]] = experiment[i]

        return {
            "best_experiment_stat": pd.DataFrame(targets_dict).T,
            "best_split_stat": experiment.iloc[-9:],
        }

    @staticmethod
    def split_splited_data(splitted_data: pd.DataFrame) -> Dict[str, pd.DataFrame]:
        """Splits a pandas DataFrame into two separate dataframes based on a specified group field.

        Args:
            splitted_data:
                The input dataframe to be split

        Returns:
            A dictionary containing two dataframes, 'test' and 'control', where 'test' contains rows where the
            group field is 'test', and 'control' contains rows where the group field is 'control'.
        """
        return {
            "control": splitted_data[splitted_data["group"] == "control"],
            "test": splitted_data[splitted_data["group"] == "test"],
        }

    def split_analysis(self, splited_data: pd.DataFrame, **kwargs):
        """Conducts a full splitting analysis.

        Args:
            splited_data:
                Data that has already been split
            **kwargs:
                Some extra keyword arguments for plots in visualization
        """
        ssp = self.split_splited_data(splited_data)
        for nf in self.target_fields:
            self.num_feature_uniform_analysis(
                ssp["control"][nf], ssp["test"][nf], **kwargs
            )
        for cf in self.group_cols:
            self.cat_feature_uniform_analysis(
                ssp["control"][cf], ssp["test"][cf], **kwargs
            )

    def get_resume(self, aa_score: pd.DataFrame, best_experiment_stat: pd.DataFrame):
        """Format results into clear format for understanding.

        Args:
            aa_score:
                Results of aa-test
            best_experiment_stat:
                Results of the best experiment

        Returns:
            DataFrame with OK and not OK depending on the results of statistical tests
        """
        result = {"aa test passed": {}, "split is uniform": {}}
        for field in self.target_fields:
            result["aa test passed"][field] = (
                aa_score.loc[field, "t-test aa passed"]
                or aa_score.loc[field, "ks-test aa passed"]
            )
            result["split is uniform"][field] = (
                best_experiment_stat.loc[field, "t-test passed"]
                or best_experiment_stat.loc[field, "ks-test passed"]
            )
        result = pd.DataFrame(result)
        result["split is uniform"] = (
            result["split is uniform"]
            .astype("bool")
            .replace({False: "OK", True: "not OK"})
        )
        result["aa test passed"] = (
            result["aa test passed"]
            .astype("bool")
            .replace({False: "not OK", True: "OK"})
        )
        return result

    def process(
        self,
        data: pd.DataFrame,
        optimize_groups: bool = False,
        iterations: int = 2000,
        show_plots: bool = True,
        test_size: float = 0.5,
        pbar: bool = True,
        **kwargs,
    ):
        """Main function for AATest estimation.

        Provides:
            * Columns labeling
            * Results calculations
            * Plotting results

        Args:
            test_size:
                Proportion of the test group
            data:
                Input dataset
            optimize_groups:
                Is in necessary to optimize groups
            iterations:
                Number of iterations for AA-test
            show_plots:
                Is in necessary to show plots
            pbar:
                Show progress-bar
            **kwargs:
                Some extra keyword arguments

        Returns:
            best_results:
                Results of the experiment with metrics for all fields
            best_split:
                Result of separation
        """
        labeling = self.columns_labeling(data)
        best_results, best_split = None, None

        if not self.target_fields:
            self.target_fields = labeling["target_field"]

        if optimize_groups:
            max_score = -1

            group_variants = [[]]
            for i in range(1, len(labeling["group_col"])):
                i_combinations = combinations(labeling["group_col"], i)
                group_variants.extend(iter(i_combinations))

            for gs in tqdm(group_variants, desc="Group optimization", disable=not pbar):
                self.group_cols = list(gs)
                experiment_results, data_splits = self.calc_uniform_tests(
                    data,
                    pbar=False,
                    iterations=iterations,
                    test_size=test_size,
                    **kwargs,
                )
                if len(experiment_results):
                    aa_scores = self.aa_score(experiment_results)
                    group_score = max(
                        aa_scores.loc["mean", "t-test aa passed"],
                        aa_scores.loc["mean", "ks-test aa passed"],
                    )
                    if group_score > max_score:
                        best_results, best_split = experiment_results, data_splits
                        max_score = group_score

        else:
            best_results, best_split = self.calc_uniform_tests(
                data,
                experiment_write_mode="full",
                split_write_mode="any",
                iterations=iterations,
                test_size=test_size,
                pbar=pbar,
                **kwargs,
            )

        if len(best_results) == 0:
            return best_results, best_split
        if len(best_results) > 0:
            if show_plots:
                aa_scores = self.uniform_tests_interpretation(best_results)
            else:
                aa_scores = self.aa_score(best_results)
            best_rs = best_results.loc[
                best_results["mean_tests_score"].idxmax(), "random_state"
            ]
            final_split = best_split[best_rs]
            if show_plots:
                self.split_analysis(final_split, **kwargs)

            transformed_results = self.experiment_result_transform(
                best_results[best_results["random_state"] == best_rs].iloc[0]
            )
            best_experiment_stat = transformed_results["best_experiment_stat"]
            best_split_stat = transformed_results["best_split_stat"]
            resume = self.get_resume(aa_scores, best_experiment_stat)
        else:
            aa_scores = None
            final_split = None
            best_experiment_stat = None
            best_split_stat = None
            resume = None

        return {
            "experiments": best_results,
            "aa_score": aa_scores,
            "split": final_split,
            "best_experiment_stat": best_experiment_stat,
            "split_stat": best_split_stat,
            "resume": resume,
        }



# Решаем задачу множественного тестирования
# Хотим выбрать лучшую выборку из k выборок по некоторой целевой метрике
# Пусть гипотеза H_i - iя выборка лучшая, i = 1, ..., k
# Пусть гипотеза H_0 - нет лучшей выборки

# Разработали метод, основанный на предельном распределении
# Он позволяет придерживаться заложенных вероятностей ошибок 1го и 2го рода

# Функции минимального размера выборки и критерия являются основными
# Функция квантиля предельного распределения случайной величины минимума используется в основных функциях


from scipy.stats import norm, bernoulli


# Решаем задачу множественного тестирования
# Хотим выбрать лучшую выборку из k выборок по некоторой целевой метрике
# Пусть гипотеза H_i - iя выборка лучшая, i = 1, ..., k
# Пусть гипотеза H_0 - нет лучшей выборки

# Разработали метод, основанный на предельном распределении
# Он позволяет придерживаться заложенных вероятностей ошибок 1го и 2го рода

# Функции минимального размера выборки и критерия являются основными
# Функция квантиля предельного распределения случайной величины минимума используется в основных функциях

def quantile_of_marginal_distribution(num_samples, quantile_level, variances=[1, 1, 1], equal_variance=True):
    """Функция квантиля предельного распределения минимума

    Parameters
    ----------
    num_samples : int
        Количество выборок, целое число больше 2
    quantile_level : float
        Уровень квантиля предельного распределения минимума
    variances : list of float, optional
        Список дисперсий выборок одинаковой длины. Количество выборок больше 2
    equal_variance : bool, optional
        Равенство дисперсий

    Returns
    -------
    float, если equal_var=True
        Квантиль предельного распределения минимума уровня gamma
    list of float, если equal_var=False
        Набор квантилей предельного распределения минимума для каждого reference_sample_index уровня gamma
    """
    iteration_size = 20000  # Количество итераций теста

    if equal_variance:
        reference_sample_index = 0  # в силу симметрии reference_sample_index по гипотезе H_0 возьмём reference_sample_index = 0 (первая выборка)
        t_values = []
        random_samples = norm.rvs(size=[iteration_size, num_samples], random_state=random_state)
        for sample in random_samples:
            min_t_value = np.inf
            for i in range(num_samples):
                if i != reference_sample_index:
                    t_value = (sample[reference_sample_index] - sample[i]) / np.sqrt(2)
                    min_t_value = min(min_t_value, t_value)
            t_values.append(min_t_value)
        return np.quantile(t_values, quantile_level)
    else:
        quantiles = []
        for j in range(num_samples):
            t_values = []
            random_samples = norm.rvs(size=[iteration_size, num_samples], random_state=random_state)
            for sample in random_samples:
                min_t_value = np.inf
                for i in range(num_samples):
                    if i != j:
                        t_value = sample[j] / np.sqrt(1 + variances[i] / variances[j]) - sample[i] / np.sqrt(
                            1 + variances[j] / variances[i])
                        min_t_value = min(min_t_value, t_value)
                t_values.append(min_t_value)
            quantiles.append(np.quantile(t_values, quantile_level))
        return quantiles


def test_on_marginal_distribution(samples, significance_level=0.05, equal_variance=True, quantiles=None):
    """Функция критерия, основанного на предельном распределении минимума

    Parameters
    ----------
    samples : list of lists
        Список выборок одинаковой длины. Количество выборок больше 2
    significance_level : float, optional
        Уровень значимости, число от 0 до 1
    equal_variance : bool, optional
        Равенство дисперсий
    quantiles : optional
            float, если equal_var=True
                Квантиль предельного распределения минимума уровня 1-alpha/len(X)
            list of float, если equal_var=False
                Набор квантилей предельного распределения минимума для каждого j уровня 1-alpha/len(X)

    Returns
    -------
    int
        Число от 0 до k - номер принятой гипотезы
    """
    num_samples = len(samples)  # Число выборок
    sample_size = len(samples[0])  # Размер выборки

    means = [np.mean(sample) for sample in samples]
    variances = [np.var(sample) * sample_size / (sample_size - 1) for sample in samples]

    if equal_variance:
        if quantiles is None:
            quantiles = quantile_of_marginal_distribution(num_samples=num_samples,
                                                          quantile_level=1 - significance_level / num_samples)  # квантиль предельного распределения
        for j in range(num_samples):
            min_t_value = np.inf
            for i in range(num_samples):
                if i != j:
                    t_value = np.sqrt(sample_size) * (means[j] - means[i]) / np.sqrt(variances[j] + variances[i])
                    min_t_value = min(min_t_value, t_value)
            if min_t_value > quantiles:
                return j + 1
        return 0
    else:
        if quantiles is None:
            quantiles = quantile_of_marginal_distribution(num_samples=num_samples,
                                                          quantile_level=1 - significance_level / num_samples,
                                                          variances=variances,
                                                          equal_variance=False)  # набор квантилей предельного распределения
        for j in range(num_samples):
            min_t_value = np.inf
            for i in range(num_samples):
                if i != j:
                    t_value = np.sqrt(sample_size) * (means[j] - means[i]) / np.sqrt(variances[j] + variances[i])
                    min_t_value = min(min_t_value, t_value)
            if min_t_value > quantiles[j]:
                return j + 1
        return 0


def min_sample_size(number_of_samples, minimum_detectable_effect, variances, significance_level=0.05, power_level=0.2,
                    equal_variance=True, quantile_1=None, quantile_2=None, initial_estimate=None):
    """Функция для подсчёта минимального размера выборки

    Parameters
    ----------
    number_of_samples : int
        Количество выборок, целое число больше 2
    minimum_detectable_effect : float
        Minimum Detectable Effect, положительное число
    variances : list of float, если equal_var=False
          float, если equal_var=True
        Оценка дисперсии выборок одинаковой длины при H(0). Количество выборок больше 2
    significance_level : float, optional
        Уровень значимости, число от 0 до 1
    power_level : float, optional
        1 - мощность, число от 0 до 1
    equal_variance : bool, optional
        Равенство дисперсий
    quantile_1 : optional
            float, если equal_var=True
                Квантиль предельного распределения минимума уровня 1-alpha/len(X)
            list of float, если equal_var=False
                Набор квантилей предельного распределения минимума для каждого sample_index уровня 1-alpha/len(X)
    quantile_2 : optional
            float, если equal_var=True
                Квантиль предельного распределения минимума уровня beta
            None, если equal_var=False
    initial_estimate : int, optional
        Нижняя граница для размера (для более быстрой работы программы)

    # В нашем случае все выборки будут одного размера

    Returns
    -------
    int
        Число sample_size - размер одной выборки
    """
    random_state = 42
    if equal_variance:
        if quantile_1 is None:
            quantile_1 = quantile_of_marginal_distribution(num_samples=number_of_samples,
                                                           quantile_level=1 - significance_level / number_of_samples)  # квантиль предельного распределения 1-alpha/k

        if quantile_2 is None:
            quantile_2 = quantile_of_marginal_distribution(num_samples=number_of_samples,
                                                           quantile_level=power_level)  # квантиль предельного распределения beta
        print(f"{quantile_1 = }, {quantile_2 = }")

        return int(2 * variances * ((quantile_1 - quantile_2) / minimum_detectable_effect) ** 2) + 1
    else:
        iteration_size = 3000  # Количество итераций
        if quantile_1 is None:
            quantile_1 = quantile_of_marginal_distribution(num_samples=number_of_samples,
                                                           quantile_level=1 - significance_level / number_of_samples,
                                                           variances=variances,
                                                           equal_variance=False)  # набор квантилей предельного распределения
        sample_sizes = []  # для размеров выборки
        for sample_index in range(number_of_samples):
            sample_size = initial_estimate or 0
            current_power = 0  # мощность
            while current_power < 1 - power_level:
                sample_size += 100
                current_power = 0
                total_samples = norm.rvs(size=[iteration_size, number_of_samples], random_state=random_state)
                for sample in total_samples:
                    min_t_value = np.inf
                    for i in range(number_of_samples):
                        if i != sample_index:
                            t_value = sample[sample_index] / np.sqrt(1 + variances[i] / variances[sample_index]) - \
                                      sample[i] / np.sqrt(
                                1 + variances[sample_index] / variances[i]) + minimum_detectable_effect * np.sqrt(
                                sample_size / (variances[sample_index] + variances[i]))
                            min_t_value = min(min_t_value, t_value)
                    if min_t_value > quantile_1[sample_index]:
                        current_power += 1
                current_power /= iteration_size
            sample_sizes.append(sample_size)
        return np.max(sample_sizes)


# Применение метода

import numpy as np

# Фиксированный random state
random_state = np.random.RandomState(42)  # Вы можете выбрать любое число в качестве seed

num_samples = 10  # число выборок
minimum_detectable_effect = 0.05  # MDE
assumed_conversion = 0.3  # предполагаемая конверсия
significance_level = 0.05  # уровень значимости
power_level = 0.2  # 1 - мощность

# Считаем минимальный размер выборки
sample_size = min_sample_size(num_samples, minimum_detectable_effect,
                              variances=assumed_conversion * (1 - assumed_conversion),
                              significance_level=significance_level, power_level=power_level, equal_variance=True)
print(f'Размер выборки = {sample_size}')
print("kek")

N = 5

# Все выборки имеют одинаковую конверсию
print('\nВсе выборки имеют одинаковую конверсию')
for _ in range(N):
    samples = bernoulli.rvs(assumed_conversion, size=[num_samples, sample_size], random_state=random_state)
    hypothesis = test_on_marginal_distribution(samples, significance_level=significance_level)
    print(f'\tПринята гипотеза H({hypothesis})')

# Десятая выборка имеет большую на MDE конверсию
print('\nДесятая выборка имеет большую на MDE конверсию')
for _ in range(N):
    samples = [bernoulli.rvs(assumed_conversion, size=sample_size, random_state=random_state) for _ in
               range(num_samples - 1)]
    samples.append(
        bernoulli.rvs(assumed_conversion + minimum_detectable_effect, size=sample_size, random_state=random_state))
    hypothesis = test_on_marginal_distribution(samples, significance_level=significance_level)
    print(f'\tПринята гипотеза H({hypothesis})')

# Пример 2
# Рассмотрим множественный тест для выявления выборки с лучшим доходом на клиента (конверсия * цена)
# В этом случае при гипотезе H_0, равенстве ARPU на всех выборках, дисперсии не равны,
# поскольку при гипотезе H_0 при разных ценах получаем разные конверсии

num_samples = 5  # число выборок
minimum_detectable_effect = 2.5  # MDE
# в среднем ARPU = 15 рублей
prices = [100, 150, 150, 200, 250]  # цены тарифов
conversions = [0.15, 0.1, 0.1, 0.075, 0.06]  # конверсии тарифов
significance_level = 0.05  # уровень значимости
power_level = 0.2  # 1 - мощность
variances = [price ** 2 * conversion * (1 - conversion) for price, conversion in zip(prices, conversions)]

# считаем минимальный размер выборки
sample_size = min_sample_size(num_samples, minimum_detectable_effect, variances=variances, significance_level=significance_level,
                              power_level=power_level, equal_variance=False)
print(f'Размер выборки = {sample_size}')

# Попробуем сгенерировать выборки и посмотреть результат тестирования
N = 5
# все выборки имеют одинаковый ARPU
print('\nВсе выборки имеют одинаковый ARPU')
for _ in range(N):
    samples = []
    for i in range(num_samples):
        samples += [prices[i] * bernoulli.rvs(conversions[i], size=sample_size)]
    hypothesis = test_on_marginal_distribution(samples, significance_level=significance_level)
    print(f'\tПринята гипотеза H({hypothesis})')

# десятая выборка имеет больший на MDE ARPU
print('\nДесятая выборка имеет больший на MDE ARPU')
for _ in range(N):
    samples = []
    for i in range(num_samples - 1):
        samples += [prices[i] * bernoulli.rvs(conversions[i], size=sample_size)]
    samples += [prices[num_samples - 1] * bernoulli.rvs(
        conversions[num_samples - 1] + minimum_detectable_effect / prices[num_samples - 1], size=sample_size)]
    hypothesis = test_on_marginal_distribution(samples, significance_level=significance_level)
    print(f'\tПринята гипотеза H({hypothesis})')



"""Calculate PSI."""
import logging

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd


logger = logging.getLogger("psi_pandas")
console_out = logging.StreamHandler()
logging.basicConfig(
    handlers=(console_out,),
    format="[%(asctime)s | %(name)s | %(levelname)s]: %(message)s",
    datefmt="%d.%m.%Y %H:%M:%S",
    level=logging.INFO,
)


class PSI:
    """Calculates population stability index for different categories of data.

    For numeric data the class generates numeric buckets, except when numeric column
    includes only NULL. For categorical data:
    1. For n < 20, a bucket equals the proportion of each category,
    2. For n > 20, a bucket equals to a group of categories,
    3. For n > 100, it calculates unique_index based on Jaccard similarity,
        but in case of imbalance null-good data returns PSI

    Args:
        expected:
            The expected values
        actual:
            The actual values
        column_name:
            The column name for which to calculate the PSI
        plot:
            If true, generates a distribution plot. Defaults to False

    Returns:
        PSI for column
        The PSI for each bucket
        New categories (empty list for non-categorical data)
        Categories that are absent in actual column (empty list for non-categorical data)
    """

    def __init__(
        self, expected: pd.DataFrame, actual: pd.DataFrame, column_name: str, plot: bool = False, silent=False
    ):
        """Initializes the PSI class with given parameters.

        Args:
            expected:
                The expected values
            actual:
                The actual values
            column_name:
                The column name for which to calculate the PSI
            plot:
                If true, generates a distribution plot. Defaults to False
            silent:
                If True show logs. Default by False
        """
        self.expected = expected[column_name].values
        self.actual = actual[column_name].values
        self.expected_len = len(self.expected)
        self.actual_len = len(self.actual)
        self.column_name = column_name
        self.column_type = self.expected.dtype
        self.expected_shape = self.expected.shape
        self.expected_nulls = np.sum(pd.isna(self.expected))
        self.actual_nulls = np.sum(pd.isna(self.actual))
        self.axis = 1
        self.plot = plot
        self.silent = silent
        if self.column_type == np.dtype("O"):
            self.expected_uniqs = expected[column_name].unique()
            self.actual_uniqs = actual[column_name].unique()

    def jac(self) -> float:
        """Calculates the Jacquard similarity index.

        The Jacquard similarity index measures the intersection between two sequences
        versus the union of the two sequences

        Returns:
            The Jacquard similarity index

        """
        x = set(self.expected_uniqs)
        y = set(self.expected_uniqs)

        logger.info(f"Jacquard similarity is {len(x.intersection(y)) / len(x.union(y)): .6f}")

        jac_sim_index = len(x.intersection(y)) / len(x.union(y))

        return jac_sim_index

    # в функции нет аргумента nulls, а был и испольовался далее в коде
    def plots(self, expected_percents, actual_percents, breakpoints, intervals):
        """Generates plots expected and actual percents.

        Args:
            expected_percents:
                The percentage of expected value from all expected values
            actual_percents:
                The percentage of actual value from all actual values
            breakpoints:
                The list of breakpoints
            intervals:
                The list of intervals

        """
        points = [i for i in breakpoints]
        plt.figure(figsize=(15, 7))
        plt.bar(
            np.arange(len(intervals)) - 0.15,  # что такое 0.15? Может вынести в константу?
            expected_percents,
            label="expected",
            alpha=0.7,
            width=0.3,
        )
        plt.bar(np.arange(len(intervals)) + 0.15, actual_percents, label="actual", alpha=0.7, width=0.3)
        plt.legend(loc="best")

        if self.column_type != np.dtype("O"):
            plt.xticks(range(len(intervals)), intervals, rotation=90)
        else:
            plt.xticks(range(len(points)), points, rotation=90)
        plt.title(self.column_name)

        # plt.savefig(f"C:\\Users\\Glazova2-YA\\Documents\\data\\bip\\summary_psi_plots\\{self.column_name}.png")
        plt.show()

    def sub_psi(self, e_perc: float, a_perc: float) -> float:
        """Calculates the sub PSI value.

        Args:
            e_perc:
                The expected percentage
            a_perc:
                The actual percentage

        Returns:
            The calculated sub PSI value.
        """
        if a_perc == 0:
            a_perc = 0.0001
        if e_perc == 0:
            e_perc = 0.0001

        sub_psi = (e_perc - a_perc) * np.log(e_perc / a_perc)

        logger.debug(f"sub_psi value is {sub_psi: .6f}")

        return sub_psi

    def psi_num(self):
        """Calculate the PSI for a single variable.

        Returns:
            PSI for column
            The PSI for each bucket
            New categories (empty list for non-categorical data)
            Categories that are absent in actual column (empty list for non-categorical data)

        """
        buckets = 10
        breakpoints = np.arange(0, buckets / 10, 0.1)

        # Заплатка, на случай, если в актуальной таблице появились значения отличные от null
        if self.expected_nulls == self.expected_len and self.actual_nulls != self.actual_len:
            breakpoints = np.array(list(sorted(set(np.nanquantile(self.actual, breakpoints)))))
        else:
            breakpoints = np.array(list(sorted(set(np.nanquantile(self.expected, breakpoints)))))

        actual_nulls = self.actual_nulls / self.actual_len
        expected_nulls = self.expected_nulls / self.expected_len

        breakpoints = np.concatenate(([-np.inf], breakpoints, [np.inf]))

        expected_percents = np.histogram(self.expected, breakpoints)
        actual_percents = np.histogram(self.actual, breakpoints)
        # breakpoints[0] = -np.inf
        # breakpoints[-1] = np.inf
        expected_percents = [p / self.expected_len for p in expected_percents[0]]
        actual_percents = [p / self.actual_len for p in actual_percents[0]]

        if self.expected_nulls == 0 and actual_nulls == expected_nulls:
            expected_percents = expected_percents
            actual_percents = actual_percents
            nulls = False
        else:
            expected_percents.append(expected_nulls)
            actual_percents.append(actual_nulls)
            nulls = True

        points = [i for i in breakpoints]
        intervals = [f"({np.round(points[i], 5)};{np.round(points[i + 1], 5)})" for i in range(len(points) - 1)]
        if nulls:
            intervals = np.append(intervals, "empty_values")

        if self.plot:
            self.plots(expected_percents, actual_percents, breakpoints, intervals)  # в функции нет аргумента nulls

        psi_dict = {}
        for i in range(0, len(expected_percents)):
            psi_val = self.sub_psi(expected_percents[i], actual_percents[i])
            psi_dict.update({intervals[i]: psi_val})

        psi_value = np.sum(list(psi_dict.values()))
        psi_dict = {k: v for k, v in sorted(psi_dict.items(), key=lambda x: x[1], reverse=True)}
        new_cats = []
        abs_cats = []

        return psi_value, psi_dict, new_cats, abs_cats

    def uniq_psi(self):
        """Calculates PSI for categorical unique counts grater than 100.

        Returns:
            PSI for column
            The PSI for each bucket
            New categories (empty list for non-categorical data)
            Categories that are absent in actual column (empty list for non-categorical data)

        """
        actual_nulls = self.actual_nulls / self.actual_len
        expected_nulls = self.expected_nulls / self.expected_len

        actual_not_nulls_arr = self.actual[~np.isnan(self.actual)]
        expected_not_nulls_arr = self.expected[~np.isnan(self.expected)]

        actual_not_nulls = len(actual_not_nulls_arr) / self.actual_len
        expected_not_nulls = len(expected_not_nulls_arr) / self.expected_len

        expected_percents = [expected_not_nulls, expected_nulls]
        actual_percents = [actual_not_nulls, actual_nulls]

        breakpoints = ["good_data", "nulls"]
        if self.plot:
            self.plots(expected_percents, actual_percents, breakpoints, breakpoints)  # в функции нет аргумента nulls

        psi_dict = {}
        for i in range(0, len(expected_percents)):
            psi_val = self.sub_psi(expected_percents[i], actual_percents[i])
            if breakpoints[i] == "None":
                psi_dict.update({"empty_value": psi_val})
            else:
                psi_dict.update({breakpoints[i]: psi_val})

        psi_value = np.sum(list(psi_dict.values()))
        jac_metric = self.jac()
        new_cats, abs_cats = [], []
        psi_dict = {k: v for k, v in sorted(psi_dict.items(), key=lambda x: x[1], reverse=True)}

        if psi_value >= 0.2:  # что такое 0.2? Может перенести его в константу?
            psi_value = psi_value
            psi_dict.update({"metric": "stability_index"})
        else:
            psi_value = 1 - jac_metric
            psi_dict.update({"metric": "unique_index"})

        logger.info(f"PSI for categorical unique >100 is {psi_value: .6f}")

        return psi_value, psi_dict, new_cats, abs_cats

    def psi_categ(self):
        """Calculates PSI for categorical data excluding unique counts grater than 100.

        Returns:
            PSI for column
            The PSI for each bucket
            New categories (empty list for non-categorical data)
            Categories that are absent in actual column (empty list for non-categorical data)

        """
        expected_uniq_count = len(self.expected_uniqs)
        actual_uniq_count = len(self.actual_uniqs)
        # правило для категориальных > 100
        if expected_uniq_count > 100 or actual_uniq_count > 100:
            psi_value, psi_dict, new_cats, abs_cats = self.uniq_psi()

            logger.info(f"PSI is {psi_value: .6f}")

            return psi_value, psi_dict, new_cats, abs_cats

        expected_dict = (
            pd.DataFrame(self.expected, columns=[self.column_name])
            .groupby(self.column_name)[self.column_name]
            .count()
            .sort_values(ascending=False)
            .to_dict()
        )
        actual_dict = (
            pd.DataFrame(self.actual, columns=[self.column_name])
            .groupby(self.column_name)[self.column_name]
            .count()
            .sort_values(ascending=False)
            .to_dict()
        )

        breakpoints = list(expected_dict.keys() | actual_dict.keys())

        new_cats = [k for k in actual_dict.keys() if k not in expected_dict.keys()]
        abs_cats = [k for k in expected_dict.keys() if k not in actual_dict.keys()]

        expected_dict_re = {}
        actual_dict_re = {}

        for b in breakpoints:
            if b in expected_dict and b not in actual_dict:
                expected_dict_re.update({b: expected_dict[b]})
                actual_dict_re.update({b: 0})
            elif b not in expected_dict and b in actual_dict:
                expected_dict_re.update({b: 0})
                actual_dict_re.update({b: actual_dict[b]})
            elif b in expected_dict and b in actual_dict:
                actual_dict_re.update({b: actual_dict[b]})
                expected_dict_re.update({b: expected_dict[b]})

        category_names = [c for c in expected_dict_re.keys()]
        groups = {}
        g_counts = len(category_names)
        group_num = 20
        if g_counts <= group_num:
            for g_n, val in enumerate(category_names):
                groups[val] = g_n
        else:
            group_size = np.floor(g_counts / group_num)
            current_pos = 0
            reminder = g_counts % group_num
            for g_n in range(group_num):
                if g_n < group_num - reminder:
                    group_values = category_names[int(current_pos) : int(current_pos + group_size)]
                    current_pos += group_size
                else:
                    group_values = category_names[int(current_pos) : int(current_pos + group_size + 1)]
                    current_pos += group_size + 1
                for val in group_values:
                    groups[val] = g_n
        group_sum_exp = 0
        group_sum_act = 0
        exp_dict = {}
        act_dict = {}
        group_re = -1
        cat_group_name = ""
        group_name_re = ""
        for k, v in groups.items():
            current_group = v
            if current_group == group_re:
                group_re = v
                exp_dict.pop(group_name_re, None)
                act_dict.pop(group_name_re, None)
                cat_group_name = cat_group_name + ", " + str(k)
                group_sum_exp += expected_dict_re[k]
                group_sum_act += actual_dict_re[k]
                exp_dict.update({cat_group_name: group_sum_exp})
                act_dict.update({cat_group_name: group_sum_act})
                group_name_re = cat_group_name
            else:
                group_name_re = str(k)
                group_re = v
                cat_group_name = str(k)
                group_sum_exp = expected_dict_re[k]
                group_sum_act = actual_dict_re[k]
                exp_dict.update({cat_group_name: group_sum_exp})
                act_dict.update({cat_group_name: group_sum_act})

        expected_percents = [e / self.expected_len for e in exp_dict.values()]
        actual_percents = [a / self.actual_len for a in act_dict.values()]

        breakpoints = [e for e in exp_dict.keys()]

        if self.plot:
            self.plots(
                expected_percents, actual_percents, breakpoints, breakpoints
            )  # в функции plots нет аргумента nulls

        psi_dict = {}
        for i in range(0, len(expected_percents)):
            psi_val = self.sub_psi(expected_percents[i], actual_percents[i])
            if breakpoints[i] == "None":
                psi_dict.update({"empty_value": psi_val})
            else:
                psi_dict.update({breakpoints[i]: psi_val})
        psi_value = np.sum(list(psi_dict.values()))
        psi_dict = {k: v for k, v in sorted(psi_dict.items(), key=lambda x: x[1], reverse=True)}

        return psi_value, psi_dict, new_cats, abs_cats

    def psi_result(self):
        """Calculates PSI.

        Returns:
            PSI for column
            The PSI for each bucket
            New categories (empty list for non-categorical data)
            Categories that are absent in actual column (empty list for non-categorical data)

        """
        if len(self.expected_shape) == 1:
            psi_values = np.empty(len(self.expected_shape))
        else:
            psi_values = np.empty(self.expected_shape[self.axis])

        for i in range(0, len(psi_values)):
            if (self.column_type == np.dtype("O")) or (
                self.expected_nulls == self.expected_len and self.actual_nulls == self.actual_len
            ):
                psi_values, psi_dict, new_cats, abs_cats = self.psi_categ()
            else:
                psi_values, psi_dict, new_cats, abs_cats = self.psi_num()

        if self.silent:
            logger.debug(f"PSI value: {psi_values: .3f}")
        else:
            logger.info(f"PSI value: {psi_values: .3f}")

        # если expected_shape пустой - будет ошибка
        return round(psi_values, 2), psi_dict, new_cats, abs_cats


def report(expected: pd.DataFrame, actual: pd.DataFrame, plot: bool = False, silent: bool = False) -> pd.DataFrame:
    """Generates a report using PSI (Population Stability Index)  between the expected and actual data.

    Args:
        expected:
            The expected dataset
        actual:
            The new dataset you want to compare to the expected one
        plot:
            If True, plots the PSI are created. Defaults to False
        silent:
            If silent, logger in info mode


    Returns:
        A dataframe with the PSI report. The report includes the columns names,
        metric names, check results, failed buckets, new categories and disappeared categories.
        Anomaly score represent the PSI, metrics names indicate with metric was used for PSI calculation,
        check results indicate whether the PSI is under the threshold (0.2),
        and failed buckets include up to 5 buckets with the highest PSI.

    """
    if silent:
        logger.debug("Creating report")
    else:
        logger.info("Creating report")

    assert len(expected.columns) == len(actual.columns)

    data_cols = expected.columns
    score_dict = {}
    new_cat_dict = {}
    datas = []

    for col in data_cols:
        psi_res = PSI(expected, actual, col, plot=plot, silent=silent)
        # отладка, в случае ошибки  выдаст прооблемный столбец
        try:
            score, psi_dict, new_cats, abs_cats = psi_res.psi_result()
        except:
            logger.warning(f"Can not count PSIs, see column {col}")
            continue

        if len(new_cats) > 0:
            new_cat_dict.update({col: new_cats})

        score_dict.update({col: score})
        check_result = "OK" if score < 0.2 else "NOK"  # может 0.2 вынести в константу?
        # psi_dict = {k:v for k,v in sorted(psi_dict.items(), key=lambda x: x[1], reverse=True)}
        failed_buckets = list(psi_dict.keys())[:5] if score > 0.2 else []
        if "metric" in psi_dict:
            new_cats = None
            abs_cats = None
            metric_name = psi_dict["metric"]
            if metric_name == "unique_index":
                failed_buckets = None
        else:
            metric_name = "stability_index"
        data_tmp = pd.DataFrame(
            {
                "column": col,
                "anomaly_score": score,
                "metric_name": metric_name,
                "check_result": check_result,
                "failed_bucket": f"{failed_buckets}",
                "new_category": f"{new_cats}",
                "disappeared_category": f"{abs_cats}",
            },
            index=[1],
        )
        datas.append(data_tmp)

    data = pd.concat(datas, ignore_index=True)

    return data



"""Function for base filtration of data."""
import pandas as pd
import numpy as np


def const_filtration(X: pd.DataFrame, threshold: float = 0.95) -> list:
    """Function removes features consist of constant value on 95%.

    Args:
        X: related dataset
        threshold: constant fill rate, default is 0.95

    Returns:
        List of filtered columns
    """
    is_const = pd.Series(0, index=X.columns, dtype=np.dtype(bool))
    for col in X.columns:
        # NaNs are not counted using unique (since np.nan != np.nan). Fill them with a unique value:
        cur_col = X.loc[:, col]
        cur_col.loc[~np.isfinite(cur_col)] = cur_col.max() + 1
        # Get values' frequency:
        freqs = cur_col.value_counts(normalize=True)
        is_const[col] = np.any(freqs > threshold)

    selected_features = ~is_const
    if np.sum(selected_features) == 0:
        raise AssertionError("All features were removed by constant filtration.")
    else:
        return X.loc[:, selected_features].columns.to_list()


def nan_filtration(X: pd.DataFrame, threshold: float = 0.8):
    """Function removes features consist of NaN value on 80%.

    Args:
        X: related dataset
        threshold: constant fill rate, default is 0.95

    Returns:
        List of filtered columns
    """
    nan_freqs = np.mean(pd.isnull(X), axis=0)
    is_sparse = nan_freqs > threshold
    selected_features = ~is_sparse
    if np.sum(selected_features) == 0:
        raise AssertionError("All features were removed by nan filtration.")
    else:
        return X.loc[:, selected_features].columns.to_list()



"""Outliers filter."""
import logging

import pandas as pd

logger = logging.getLogger("outliers_filter")
console_out = logging.StreamHandler()
logging.basicConfig(
    handlers=(console_out,),
    format="[%(asctime)s | %(name)s | %(levelname)s]: %(message)s",
    datefmt="%d.%m.%Y %H:%M:%S",
    level=logging.INFO,
)


class OutliersFilter:
    """Class of Outliers Filter. It creates a row indices that should be deleted by percentile."""

    def __init__(self, interquartile_coeff, mode_percentile, min_percentile, max_percentile):
        """Initializes the OutliersFilter.

        Args:
            interquartile_coeff:
                Coefficient for the interquartile range to determine outliers
            mode_percentile:
                If True, outliers are determined by custom percentiles
            min_percentile:
                The lower percentile. Values below this percentile are considered outliers.
            max_percentile:
                The upper percentile. Values above this percentile are considered outliers
        """
        self.interquartile_coeff = interquartile_coeff
        self.mode_percentile = mode_percentile
        self.min_percentile = min_percentile
        self.max_percentile = max_percentile

    def perform_filter(self, df: pd.DataFrame, interquartile: bool = True) -> pd.DataFrame:
        """Identifies rows with outliers.

        This method creates a set of row indices to be removed, which contains values less than
        `min_percentile` and larger than `max_percentile` (if `mode_percentile` is True), or values
        smaller than the 0.2 and larget than 0.8 (if `mode_percentile` is False)

        Args:
            df:
                The input DataFrame
            interquartile:
                If True, uses the interquartile range to determine outliers. Defaults to True

        Returns:
            The set of row indices with outliers
        """
        columns_names = df.select_dtypes(include="number").columns
        rows_for_del = []
        for column in columns_names:
            if self.mode_percentile:
                min_value = df[column].quantile(self.min_percentile)
                max_value = df[column].quantile(self.max_percentile)
            elif interquartile:
                upper_quantile = df[column].quantile(0.8)
                lower_quantile = df[column].quantile(0.2)

                interquartile_range = upper_quantile - lower_quantile
                min_value = lower_quantile - self.interquartile_coeff * interquartile_range
                max_value = upper_quantile + self.interquartile_coeff * interquartile_range
            else:
                mean_value = df[column].mean()
                standard_deviation = df[column].std()
                nstd_lower, nstd_upper = 3, 3

                min_value = mean_value - nstd_lower * standard_deviation
                max_value = mean_value + nstd_upper * standard_deviation

            rows_for_del_column = (df[column] < min_value) | (df[column] > max_value)
            rows_for_del_column = df.index[rows_for_del_column].tolist()
            rows_for_del.extend(rows_for_del_column)
        rows_for_del = list(set(rows_for_del))
        logger.info(f"Drop {len(rows_for_del)} rows")
        return df.drop(rows_for_del)



"""Here is the realization of Matching without replacement."""
from typing import Tuple

import numpy as np
import pandas as pd
from scipy.optimize import linear_sum_assignment
from scipy.spatial import distance

from .faiss_matcher import conditional_covariance


class MatcherNoReplacement:
    """Matching groups with no replacement.

    Realized by optimizing the linear sum of distances between pairs of treatment and
    control samples.
    """

    def __init__(self, X: pd.DataFrame, a: pd.Series, weights: dict = None, approximate_match: bool = False):
        """Initialize matching.

        Args:
            X: features dataframe
            a: series of treatment value
            weights: weights for numeric columns in order to increase matching quality.
            approximate_match: use or not approximate matching
        """
        self.treatment = a
        self.X = X
        self.weights = weights
        self.approximate_match = approximate_match

    def match(self):
        """Function run matching with no replacement.

        Returns:
            Dataframe of matched indexes.
        """
        matches = {}
        cov = conditional_covariance(self.X[self.treatment == 1].values, self.X[self.treatment == 0].values)
        distance_matrix = self._get_distance_matrix(self.X[self.treatment == 1], self.X[self.treatment == 0], cov)
        source_array, neighbor_array_indices, distances = optimally_match_distance_matrix(distance_matrix)
        source_df = self.X[self.treatment == 1].iloc[np.array(source_array)]
        target_df = self.X[self.treatment == 0].iloc[np.array(neighbor_array_indices)]

        matches[1] = self.create_match_df(self.treatment, source_df, target_df, distances)
        matches[0] = self.create_match_df(self.treatment, target_df, source_df, distances)

        match_df = pd.concat(matches, sort=True)
        return match_df

    def create_match_df(
            self, base_series: pd.Series, source_df: pd.DataFrame, target_df: pd.DataFrame, distances: list
    ) -> pd.DataFrame:
        """Function creates matching dataframe.

        Args:
            base_series: series of treatment value.
            source_df: dataframe of sources indexes.
            target_df: dataframe of target indexes.
            distances: matrix of calculated distances.

        Returns:
            Matched dataframe of indexes.
        """
        match_sub_df = pd.DataFrame(
            index=base_series.index,
            columns=[
                "matches",
                "distances",
            ],
            data=base_series.apply(lambda x: pd.Series([[], []])).values,
            dtype="object",
        )

        # matching from source to target: read distances
        match_sub_df.loc[source_df.index] = pd.DataFrame(
            data=dict(
                matches=[[tidx] for tidx in target_df.index],
                distances=distances,
            ),
            index=source_df.index,
        )

        # matching from target to target: fill with zeros
        match_sub_df.loc[target_df.index] = pd.DataFrame(
            data=dict(
                matches=[[tidx] for tidx in target_df.index],
                distances=[[0]] * len(distances),
            ),
            index=target_df.index,
        )
        return match_sub_df

    def _get_metric_dict(self, cov: np.ndarray) -> dict:
        """Function calculates correct feature space and generate metrics dist for cdist calculation.

        Args:
            cov: Matrix of covariations.

        Returns:
            Metric dictionary
        """
        metric_dict = dict(metric="mahalanobis")
        mahalanobis_transform = np.linalg.inv(cov)
        if self.weights is not None:
            features = self.X.columns
            w_list = np.array([self.weights[col] if col in self.weights.keys() else 1 for col in features])
            w_matrix = np.sqrt(np.diag(w_list / w_list.sum()))
            mahalanobis_transform = np.dot(w_matrix, mahalanobis_transform)

        metric_dict["VI"] = mahalanobis_transform
        return metric_dict

    def _get_distance_matrix(self, source_df: pd.DataFrame, target_df: pd.DataFrame, cov: np.ndarray) -> np.ndarray:
        """Create distance matrix for no replacement match.

        Combines metric and source/target data into a
        precalculated distance matrix which can be passed to
        scipy.optimize.linear_sum_assignment.

        Args:
            source_df: source feature dataframe.
            target_df: target feature dataframe.
            cov: matrix of covariations.

        Returns:
            Matrix of distances.
        """
        cdist_args = dict(XA=_ensure_array_columnlike(source_df.values), XB=_ensure_array_columnlike(target_df.values))
        cdist_args.update(self._get_metric_dict(cov))

        if self.approximate_match:
            if len(cdist_args['XB']) < len(cdist_args['XA']):
                covariance_matrix = np.cov(cdist_args['XB'].T)
            else:
                covariance_matrix = np.cov(cdist_args['XA'].T)
            covariance_matrix_reg = covariance_matrix + np.eye(covariance_matrix.shape[0]) * 1e-8

            distance_matrix = np.zeros((cdist_args['XA'].shape[0], cdist_args['XB'].shape[0]))
            for i, x in enumerate(cdist_args['XA']):
                distance_matrix[i] = _m_distance(cdist_args['XB'], x, np.linalg.inv(covariance_matrix_reg))
        else:
            distance_matrix = distance.cdist(**cdist_args)
        return distance_matrix


def optimally_match_distance_matrix(distance_matrix: np.ndarray) -> Tuple[np.ndarray, np.ndarray, list]:
    """Functions finds optimal neighbor with no replacement.

    Args:
        distance_matrix: matrix of distances.

    Returns:
        - indexes of source dataframe.
        - optimal neighbors array for source array.
        - distances of optimal neighbors.
    """
    source_array, neighbor_array_indices = linear_sum_assignment(distance_matrix)
    distances = [[distance_matrix[s_idx, t_idx]] for s_idx, t_idx in zip(source_array, neighbor_array_indices)]
    return source_array, neighbor_array_indices, distances


def _ensure_array_columnlike(target_array: np.ndarray) -> np.ndarray:
    """Function checks if array is column like and reshape it in order it is not.

    Args:
        target_array: checked array.

    Returns:
        column like target array.
    """
    if len(target_array.shape) < 2 or target_array.shape[1] == 1:
        target_array = target_array.reshape(-1, 1)
    return target_array


def _m_distance(X: np.ndarray, Y: np.ndarray, inv_covariance: np.ndarray) -> np.ndarray:
    """Calculation of the approximate Mahalanobis distance.
    Args:
        X: source feature dataframe.
        Y: iterable rows for calculating distance between X and Y.
        inv_covariance: inverted matrix of covariations.

    Returns:
        Mahalanobis distance between X and Y."""

    diff = X - Y
    return np.sqrt(np.sum(np.dot(diff, inv_covariance) * diff, axis=1))
